# Copyright 2023 Google LLC

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     https://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for time_integration."""

import functools
import math

from absl.testing import absltest
from absl.testing import parameterized

from dinosaur import time_integration
import jax
from jax import config
from jax import tree_util

import jax.numpy as jnp
import numpy as np

config.update('jax_enable_x64', True)


def assert_allclose(x, y, rtol=1e-7, atol=0, err_msg=''):
  jax.tree_util.tree_map(
      functools.partial(
          np.testing.assert_allclose, atol=atol, rtol=rtol, err_msg=err_msg),
      x, y)


def harmonic_oscillator(x0, t):
  theta = jnp.arctan(x0[0] / x0[1])
  r = jnp.linalg.norm(x0, ord=2, axis=0)
  return r * jnp.stack([jnp.sin(t + theta), jnp.cos(t + theta)])


class CustomODE(time_integration.ImplicitExplicitODE):

  def __init__(self, explicit_terms, implicit_terms, implicit_inverse):
    self.explicit_terms = explicit_terms
    self.implicit_terms = implicit_terms
    self.implicit_inverse = implicit_inverse


class CustomExplicitODE(time_integration.ExplicitODE):

  def __init__(self, explicit_terms):
    self.explicit_terms = explicit_terms


ALL_TEST_PROBLEMS = [
    # x(t) = np.ones(10)
    dict(testcase_name='_zero_derivative',
         explicit_terms=lambda x: 0 * x,
         implicit_terms=lambda x: 0 * x,
         implicit_inverse=lambda x, eta: x,
         dt=1e-2,
         inner_steps=10,
         outer_steps=5,
         initial_state=np.ones(10),
         closed_form=lambda x0, t: x0,
         tolerances=[1e-12] * 6),
    # x(t) = 5 * t * np.ones(3)
    dict(testcase_name='_constant_derivative',
         explicit_terms=lambda x: 5 * jnp.ones_like(x),
         implicit_terms=lambda x: 0 * x,
         implicit_inverse=lambda x, eta: x,
         dt=1e-2,
         inner_steps=10,
         outer_steps=5,
         initial_state=np.ones(3),
         closed_form=lambda x0, t: x0 + 5 * t,
         tolerances=[1e-12] * 6),
    # x(t) = np.arange(3) * np.exp(t)
    # Uses explicit terms only.
    dict(testcase_name='_linear_derivative_explicit',
         explicit_terms=lambda x: x,
         implicit_terms=lambda x: 0 * x,
         implicit_inverse=lambda x, eta: x,
         dt=1e-2,
         inner_steps=20,
         outer_steps=5,
         initial_state=np.arange(3.0),
         closed_form=lambda x0, t: np.arange(3) * jnp.exp(t),
         tolerances=[5e-2, 1e-4, 1e-6, 1e-9, 1e-6, 5e-2]),
    # x(t) = np.arange(3) * np.exp(t)
    # Uses implicit terms only.
    dict(testcase_name='_linear_derivative_implicit',
         explicit_terms=lambda x: 0 * x,
         implicit_terms=lambda x: x,
         implicit_inverse=lambda x, eta: x / (1 - eta),
         dt=1e-2,
         inner_steps=20,
         outer_steps=5,
         initial_state=np.arange(3.0),
         closed_form=lambda x0, t: np.arange(3) * jnp.exp(t),
         tolerances=[5e-2, 5e-5, 1e-5, 1e-5, 3e-5, 5e-2]),
    # x(t) = np.arange(3) * np.exp(t)
    # Splits the equation into an implicit and explicit term.
    dict(testcase_name='_linear_derivative_semi_implicit',
         explicit_terms=lambda x: x / 2,
         implicit_terms=lambda x: x / 2,
         implicit_inverse=lambda x, eta: x / (1 - eta / 2),
         dt=1e-2,
         inner_steps=20,
         outer_steps=5,
         initial_state=np.arange(3) * np.exp(0),
         closed_form=lambda x0, t: np.arange(3.0) * jnp.exp(t),
         tolerances=[1e-4, 2e-5, 2e-6, 1e-6, 2e-5, 1e-4]),
    dict(testcase_name='_harmonic_oscillator_explicit',
         explicit_terms=lambda x: jnp.stack([x[1], -x[0]]),
         implicit_terms=jnp.zeros_like,
         implicit_inverse=lambda x, eta: x,
         dt=1e-2,
         inner_steps=20,
         outer_steps=5,
         initial_state=np.ones(2),
         closed_form=harmonic_oscillator,
         tolerances=[1e-2, 3e-5, 6e-8, 5e-11, 6e-8, 1e-2]),
    dict(testcase_name='_harmonic_oscillator_implicit',
         explicit_terms=jnp.zeros_like,
         implicit_terms=lambda x: jnp.stack([x[1], -x[0]]),
         implicit_inverse=lambda x, eta: jnp.stack(  # pylint: disable=g-long-lambda
             [x[0] + eta * x[1], x[1] - eta * x[0]]) / (1 + eta ** 2),
         dt=1e-2,
         inner_steps=20,
         outer_steps=5,
         initial_state=np.ones(2),
         closed_form=harmonic_oscillator,
         tolerances=[1e-2, 2e-5, 2e-6, 1e-6, 6e-6, 1e-2]),
]


ALL_TIME_STEPPERS = [
    time_integration.backward_forward_euler,
    time_integration.crank_nicolson_rk2,
    time_integration.crank_nicolson_rk3,
    time_integration.crank_nicolson_rk4,
    time_integration.imex_rk_sil3,
    time_integration.semi_implicit_leapfrog,
]


class TimeIntegrationTest(parameterized.TestCase):

  @parameterized.named_parameters(ALL_TEST_PROBLEMS)
  def test_implicit_inverse(
      self,
      explicit_terms,
      implicit_terms,
      implicit_inverse,
      dt,
      inner_steps,
      outer_steps,
      initial_state,
      closed_form,
      tolerances,
  ):
    """Tests that time integration is accurate for a range of test cases."""
    del dt, explicit_terms, inner_steps, outer_steps, closed_form  # unused
    del tolerances  # unused

    # Verifies that `implicit_inverse` solves (y - eta * F(y)) = x
    # This does not test the integrator, but rather verifies that the test
    # case is valid.
    eta = 0.3
    solved_state = implicit_inverse(initial_state, eta)
    reconstructed_state = solved_state - eta * implicit_terms(solved_state)
    np.testing.assert_allclose(reconstructed_state, initial_state)

  @parameterized.named_parameters(ALL_TEST_PROBLEMS)
  def test_integration(
      self,
      explicit_terms,
      implicit_terms,
      implicit_inverse,
      dt,
      inner_steps,
      outer_steps,
      initial_state,
      closed_form,
      tolerances,
  ):
    # Compute closed-form solution.
    time = dt * inner_steps * (1 + np.arange(outer_steps))
    expected = jax.vmap(closed_form, in_axes=(None, 0))(
        initial_state, time)

    # Compute trajectory using time-stepper.
    for atol, time_stepper in zip(tolerances, ALL_TIME_STEPPERS):
      with self.subTest(time_stepper.__name__):
        equation = CustomODE(explicit_terms, implicit_terms, implicit_inverse)
        semi_implicit_step = time_stepper(equation, dt)
        if time_stepper == time_integration.semi_implicit_leapfrog:
          input_state = (closed_form(initial_state, 0 * dt),
                         closed_form(initial_state, 1 * dt))
          post_process_fn = lambda x: x[0]
          trajectory_fn = time_integration.trajectory_from_step(
              semi_implicit_step, outer_steps, inner_steps,
              post_process_fn=post_process_fn)
        else:
          input_state = initial_state
          trajectory_fn = time_integration.trajectory_from_step(
              semi_implicit_step, outer_steps, inner_steps)
        _, actual = trajectory_fn(input_state)
        np.testing.assert_allclose(expected, actual, atol=atol, rtol=0)

  def test_pytree_state(self):
    equation = CustomODE(
        explicit_terms=lambda x: tree_util.tree_map(jnp.zeros_like, x),
        implicit_terms=lambda x: tree_util.tree_map(jnp.zeros_like, x),
        implicit_inverse=lambda x, eta: x,
    )
    u0 = {'x': 1.0, 'y': 1.0}
    for time_stepper in ALL_TIME_STEPPERS:
      with self.subTest(time_stepper.__name__):
        if time_stepper != time_integration.semi_implicit_leapfrog:
          u1 = time_stepper(equation, 1.0)(u0)
          self.assertEqual(u0, u1)
        else:
          u0_leapfrog = (u0, u0)
          u1_leapfrog = time_stepper(equation, 1.0)(u0_leapfrog)
          self.assertEqual(u0_leapfrog[0], u1_leapfrog[0])
          self.assertEqual(u0_leapfrog[1], u1_leapfrog[1])

  def test_multiple_equations(self):
    tolerances = [1e-4, 2e-5, 2e-6, 1e-6, 2e-5, 1e-4]
    dt = 1e-2
    inner_steps = 20
    outer_steps = 5
    with self.subTest('array_inputs'):
      initial_state = np.arange(3) * np.exp(0)
      closed_form = lambda x0, t: np.arange(3.0) * jnp.exp(t)
      equation_a = CustomODE(
          explicit_terms=lambda x: 3 * x / 8,
          implicit_terms=lambda x: x / 2,
          implicit_inverse=lambda x, eta: x / (1 - eta / 2),
      )
      equation_b = CustomExplicitODE(explicit_terms=lambda x: x / 8)
      equation = time_integration.compose_equations([equation_a, equation_b])
      # Compute closed-form solution.
      time = dt * inner_steps * (1 + np.arange(outer_steps))
      expected = jax.vmap(closed_form, in_axes=(None, 0))(
          initial_state, time)

      # Compute trajectory using time-stepper.
      for atol, time_stepper in zip(tolerances, ALL_TIME_STEPPERS):
        with self.subTest(time_stepper.__name__):
          semi_implicit_step = time_stepper(equation, dt)
          if time_stepper == time_integration.semi_implicit_leapfrog:
            input_state = (closed_form(initial_state, 0 * dt),
                           closed_form(initial_state, 1 * dt))
            post_process_fn = lambda x: x[0]
            trajectory_fn = time_integration.trajectory_from_step(
                semi_implicit_step, outer_steps, inner_steps,
                post_process_fn=post_process_fn)
          else:
            input_state = initial_state
            trajectory_fn = time_integration.trajectory_from_step(
                semi_implicit_step, outer_steps, inner_steps)
          _, actual = trajectory_fn(input_state)
          np.testing.assert_allclose(expected, actual, atol=atol, rtol=0)

    with self.subTest('pytree_inputs'):
      initial_state = {'s': np.arange(3) * np.exp(0)}
      closed_form = lambda x0, t: {'s': np.arange(3.0) * jnp.exp(t)}
      equation_a = CustomODE(
          explicit_terms=lambda x: {'s': x['s'] / 8},
          implicit_terms=lambda x: {'s': x['s'] / 2},
          implicit_inverse=lambda x, eta: {'s': x['s'] / (1 - eta / 2)},
      )
      equation_b = CustomExplicitODE(
          explicit_terms=lambda x: {'s': 3 * x['s'] / 8},
      )
      equation = time_integration.compose_equations([equation_a, equation_b])
      # Compute closed-form solution.
      time = dt * inner_steps * (1 + np.arange(outer_steps))
      expected = jax.vmap(closed_form, in_axes=(None, 0))(
          initial_state, time)

      # Compute trajectory using time-stepper.
      for atol, time_stepper in zip(tolerances, ALL_TIME_STEPPERS):
        with self.subTest(time_stepper.__name__):
          semi_implicit_step = time_stepper(equation, dt)
          if time_stepper == time_integration.semi_implicit_leapfrog:
            input_state = (closed_form(initial_state, 0 * dt),
                           closed_form(initial_state, 1 * dt))
            post_process_fn = lambda x: x[0]
            trajectory_fn = time_integration.trajectory_from_step(
                semi_implicit_step, outer_steps, inner_steps,
                post_process_fn=post_process_fn)
          else:
            input_state = initial_state
            trajectory_fn = time_integration.trajectory_from_step(
                semi_implicit_step, outer_steps, inner_steps)
          _, actual = trajectory_fn(input_state)
          np.testing.assert_allclose(expected['s'], actual['s'], atol=atol)


class NestedCheckpointScanTest(parameterized.TestCase):

  @parameterized.parameters(
      dict(f=lambda x, _: (x + 1, x ** 2), init=1, xs=None,
           lengths=[8]),
      dict(f=lambda x, _: (x + 1, x ** 2), init=1, xs=None,
           lengths=[2, 4]),
      dict(f=lambda x, _: (x + 1, x ** 2), init=1, xs=None,
           lengths=[1, 2, 2, 2]),
      dict(f=lambda x, y: (2 * x, x + y), init=1, xs=np.arange(8),
           lengths=[8]),
      dict(f=lambda x, y: (2 * x, x + y), init=1, xs=np.arange(8),
           lengths=[2, 4]),
      dict(f=lambda x, y: (2 * x, x + y), init=1, xs=np.arange(8),
           lengths=[2, 2, 2, 1]),
  )
  def test(self, f, init, xs, lengths):
    length = math.prod(lengths)
    expected_c, expected_ys = jax.lax.scan(f, init, xs, length)
    actual_c, actual_ys = time_integration.nested_checkpoint_scan(
        f, init, xs, length, nested_lengths=lengths)
    np.testing.assert_allclose(expected_c, actual_c)
    np.testing.assert_allclose(expected_ys, actual_ys)


class DigitalFilterInitiailizationTest(absltest.TestCase):

  def test_accumulate_repeated(self):
    result = time_integration.accumulate_repeated(
        lambda x: 2*x, jnp.arange(4), 1)
    self.assertEqual(result, 0*2 + 1*4 + 2*8 + 3*16)

  def test_dfi_lanczos_weights(self):
    weights = time_integration._dfi_lanczos_weights(10, 10, 0.1)
    self.assertEqual(weights.size, 50)
    self.assertTrue((np.diff(weights) < 0).all())
    self.assertAlmostEqual(weights[0], 1.0, places=2)
    self.assertAlmostEqual(weights[-1], 0.0, places=2)

  def test_digital_filter_initialization(self):
    eq = time_integration.ImplicitExplicitODE.from_functions(
        # x[0] is linear growth
        # x[1] and x[2] are a simple harmonic oscillator
        explicit_terms=lambda x: jnp.stack([0.1, x[2], -x[1]]),
        implicit_terms=jnp.zeros_like,
        implicit_inverse=lambda x, eta: x,
    )
    dfi = time_integration.digital_filter_initialization(
        equation=eq,
        ode_solver=time_integration.imex_rk_sil3,
        filters=[],
        time_span=20.0,
        cutoff_period=20.0,
        dt=0.01,
    )
    result = dfi(np.array([1.0, 1.0, 1.0]))
    expected = np.array([1.0, 0.0, 0.0])  # oscillating terms are filtered
    np.testing.assert_allclose(expected, result, atol=1e-3)


if __name__ == '__main__':
  absltest.main()



# Copyright 2023 Google LLC

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     https://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for vertical_interpolation."""

from absl.testing import absltest
from absl.testing import parameterized
from dinosaur import sigma_coordinates
from dinosaur import vertical_interpolation
import jax.numpy as jnp
import numpy as np


class InterpTest(parameterized.TestCase):

  @parameterized.parameters(
      dict(x=0.5, expected=10.0),
      dict(x=1.1, expected=11.0),
      dict(x=2.5, expected=22.5),
      dict(x=10.0, expected=30.0),
  )
  def test_interp(self, x, expected):
    xp = jnp.array([1, 2, 4])
    fp = jnp.array([10, 20, 30])
    actual = vertical_interpolation.interp(x, xp, fp)
    self.assertAlmostEqual(actual, expected)

  @parameterized.parameters(
      dict(x=0.5, expected=5.0),
      dict(x=1.1, expected=11.0),
      dict(x=2.5, expected=22.5),
      dict(x=6.0, expected=40.0),
  )
  def test_linear_interp_with_linear_extrap(self, x, expected):
    xp = jnp.array([1, 2, 4])
    fp = jnp.array([10, 20, 30])
    actual = vertical_interpolation.linear_interp_with_linear_extrap(x, xp, fp)
    self.assertAlmostEqual(actual, expected)


class PressureLevelsTest(parameterized.TestCase):

  def test_get_surface_pressure(self):
    levels = np.array([100, 200, 300, 400, 500])
    orography = np.array([[0, 5, 10, 15]])
    geopotential = np.moveaxis(
        [[
            [400, 250, 150, 50, -50],
            [1000, 900, 140, 40, 20],
            [500, 400, 300, 200, 100],
            [600, 500, 400, 300, 200],
        ]],
        -1,
        0,
    )  # reorder from [x, y, level] to [level, x, y]
    expected = np.array([[[450, 390, 500, 550]]])
    actual = vertical_interpolation.get_surface_pressure(
        vertical_interpolation.PressureCoordinates(levels),
        geopotential,
        orography,
        gravity_acceleration=10,
    )
    np.testing.assert_allclose(actual, expected)

  def test_sigma_to_pressure_roundtrip(self):
    sigma_coords = sigma_coordinates.SigmaCoordinates.equidistant(10)
    pressure_coords = vertical_interpolation.PressureCoordinates(
        np.array([100, 200, 300, 400])
    )
    surface_pressure = np.array([[[250, 350, 450]]])
    original = np.moveaxis([[[1, 2, 3, 4]] * 3], -1, 0)
    on_sigma_levels = vertical_interpolation.interp_pressure_to_sigma(
        original,
        pressure_coords,
        sigma_coords,
        surface_pressure,
    )
    self.assertEqual(on_sigma_levels.shape, (10, 1, 3))
    roundtripped = vertical_interpolation.interp_sigma_to_pressure(
        on_sigma_levels, pressure_coords, sigma_coords, surface_pressure
    )
    # mask points below the surface
    expected = original.astype(float)
    expected[2:, :, 0] = np.nan
    expected[3:, :, 1] = np.nan
    np.testing.assert_allclose(roundtripped, expected, atol=1e-6)


class HybridCoordinatesTest(absltest.TestCase):

  def test_ecmwf_137(self):
    coords = vertical_interpolation.HybridCoordinates.ECMWF137()
    self.assertEqual(coords.layers, 137)

    # spot check ph (half-level, on the boundary) from row 84 in the table
    expected = 316.3607
    actual = coords.a_boundaries[84] + coords.b_boundaries[84] * 1013.250
    self.assertAlmostEqual(expected, actual, places=3)

  def test_ufs_127(self):
    coords = vertical_interpolation.HybridCoordinates.UFS127()
    self.assertEqual(coords.layers, 127)

    # spot check pf (full-level, centered) from row 80 in the table
    expected = 572.829345703125
    p_below = coords.a_boundaries[79] + coords.b_boundaries[79] * 1000
    p_above = coords.a_boundaries[80] + coords.b_boundaries[80] * 1000
    actual = (p_below + p_above) / 2
    self.assertAlmostEqual(expected, actual, places=1)

  def test_to_approx_sigma_coords(self):
    hybrid_coords = vertical_interpolation.HybridCoordinates(
        # pressure bounds at [0, 500, 800, 1000]
        a_boundaries=np.array([0, 500, 300, 0]),
        b_boundaries=np.array([0, 0, 0.5, 1.0]),
    )

    expected = np.array([0.0, 0.65, 1.0])
    sigma_coords = hybrid_coords.to_approx_sigma_coords(
        surface_pressure=1000, layers=2
    )
    np.testing.assert_allclose(sigma_coords.boundaries, expected)

    expected = np.array([0.0, 0.5, 0.8, 1.0])
    sigma_coords = hybrid_coords.to_approx_sigma_coords(
        surface_pressure=1000, layers=3
    )
    np.testing.assert_allclose(sigma_coords.boundaries, expected)

    expected = np.array([0.0, 0.25, 0.5, 0.65, 0.8, 0.9, 1.0])
    sigma_coords = hybrid_coords.to_approx_sigma_coords(
        surface_pressure=1000, layers=6
    )
    np.testing.assert_allclose(sigma_coords.boundaries, expected)

  def test_interp_hybrid_to_sigma(self):
    sigma_coords = sigma_coordinates.SigmaCoordinates.equidistant(5)
    hybrid_coords = vertical_interpolation.HybridCoordinates(
        # pressure bounds at [0, 400, 800, 1000] -> centers at [200, 600, 900]
        a_boundaries=np.array([0, 400, 300, 0]),
        b_boundaries=np.array([0, 0, 0.5, 1.0]),
    )
    surface_pressure = np.array([[1000]])
    original = np.array([1.0, 2.0, 3.0])[:, np.newaxis, np.newaxis]
    # linear interpolation (with extrapolation) from y=[1, 2, 3] defined
    # at x=[200, 600, 900] to x=[100, 300, 500, 700, 900].
    expected = np.array([0.75, 1.25, 1.75, 2 + 1 / 3, 3.0])
    actual = vertical_interpolation.interp_hybrid_to_sigma(
        original, hybrid_coords, sigma_coords, surface_pressure
    ).ravel()
    np.testing.assert_allclose(actual, expected, atol=1e-6)

  def test_regrid_hybrid_to_sigma(self):
    sigma_coords = sigma_coordinates.SigmaCoordinates.equidistant(5)
    hybrid_coords = vertical_interpolation.HybridCoordinates(
        # at pressures boundaries [0, 30, 75, 200, 450, 700, 850, 1000]
        a_boundaries=np.array([0, 30, 75, 200, 300, 300, 150, 0]),
        b_boundaries=np.array([0, 0, 0, 0, 0.15, 0.4, 0.7, 1]),
    )
    surface_pressure = np.array([[1000]])
    original = np.arange(1.0, 8.0)[:, np.newaxis, np.newaxis]
    # area weighted averages for cells centered at [100, 300, 500, 700, 900]
    expected = np.array([
        1 * (30 / 200) + 2.0 * (45 / 200) + 3.0 * (125 / 200),
        4.0,
        4.0 * (50 / 200) + 5.0 * (150 / 200),
        5.0 * (100 / 200) + 6.0 * (100 / 200),
        6.0 * (50 / 200) + 7.0 * (150 / 200),
    ])
    actual = vertical_interpolation.regrid_hybrid_to_sigma(
        original, hybrid_coords, sigma_coords, surface_pressure
    ).ravel()
    np.testing.assert_allclose(actual, expected, atol=1e-6)

  def test_interval_overlap(self):
    actual = vertical_interpolation._interval_overlap(
        np.array([1, 4, 6, 9, 10]), np.array([0, 3, 7, 10])
    )
    expected = np.array([[2, 0, 0, 0], [1, 2, 1, 0], [0, 0, 2, 1]])
    np.testing.assert_array_equal(expected, actual)


if __name__ == '__main__':
  absltest.main()



# Copyright 2023 Google LLC

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     https://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Filters for GCM models."""
import functools
from typing import Callable

from dinosaur import spherical_harmonic
from dinosaur import typing

import jax
import jax.numpy as jnp
import numpy as np


def _preserves_shape(target, scaling):
  target_shape = np.shape(target)
  return target_shape == np.broadcast_shapes(target_shape, scaling.shape)


def _make_filter_fn(scaling, name=None):
  rescale = lambda x: scaling * x if _preserves_shape(x, scaling) else x
  return functools.partial(
      jax.tree_util.tree_map, jax.named_call(rescale, name=name))


def exponential_filter(
    grid: spherical_harmonic.Grid,
    attenuation: float | typing.Array = 16,
    order: int | typing.Array = 18,
    cutoff: float = 0,
) -> Callable[[typing.PyTreeState], typing.PyTreeState]:
  """Returns a filter that attenuates modes with high total wavenumber.

  Components with `k > cutoff` are damped by a factor of:

    exp(-attenuation * ((k - cutoff) / (1 - cutoff)) ** (2 * order))

  where `k = total_wavenumber / maximum_total_wavenumber`.

  For more on the theory of filtering for resolving Gibbs Phenomenon, see Ref
  [1]. Default parameters here are chosen here based on Ref [2], with the
  attenuation factor adjusted for float32 precision.

  Args:
    grid: the `spherical_harmonic.Grid` to use for the computation.
    attenuation: controls the steepness of the attenuation above the cutoff
      frequency. Typically attentuation is chosen as -log(epsilon), so the max
      frequency components are multiplied by floating point esilon. Using
      `Array` specifications enable applying variable filter parameters for
      different levels/times.
    order: controls the polynomial order of the exponential filter. A higher
      order filter is smoother, and starts attenuating at a higher frequency.
    cutoff: a hard threshold with which to start attenuation, expressed as a
      proportion of maximum total wavenumber, e.g., as used in Reference [3].

  Returns:
    A function that accepts a state and returns a filtered state.

  References:
  [1] Gottlieb, D. & Shu, C.-W. On the Gibbs Phenomenon and Its Resolution. SIAM
      Rev. 39, 644-668 (1997)
  [2] Hou, T. Y. & Li, R. Computing nearly singular solutions using
      pseudo-spectral methods. J. Comput. Phys. 226, 379-397 (2007)
  [3] Arbic, B. K. & Flierl, G. R. Coherent vortices and kinetic energy ribbons
      in asymptotic, quasi two-dimensional f-plane turbulence. Phys. Fluids 15,
      2177-2189 (2003)
  """

  _, total_wavenumber = grid.modal_axes

  k = total_wavenumber / total_wavenumber.max()
  a = attenuation
  c = cutoff
  p = order

  scaling = jnp.exp((k > c) * (-a * (((k - c) / (1 - c)) ** (2 * p))))
  return _make_filter_fn(scaling, "exponential_filter")


def horizontal_diffusion_filter(
    grid: spherical_harmonic.Grid,
    scale: float | typing.Array,
    order: int = 1,
) -> Callable[[typing.PyTreeState], typing.PyTreeState]:
  """Returns a filter that applies a horizontal diffusion step."""
  eigenvalues = grid.laplacian_eigenvalues
  scaling = jnp.exp(-scale * (-eigenvalues) ** order)
  return _make_filter_fn(scaling, "horizontal_diffusion_filter")



# Copyright 2023 Google LLC

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     https://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""jax.numpy functions that could belong upstream."""
from __future__ import annotations

import functools
import math
import re

import jax
from jax import lax
from jax.experimental import shard_map
import jax.numpy as jnp
import numpy as np


# pylint: disable=logging-fstring-interpolation


@jax.named_call
def _single_device_dot_cumsum(
    x: jax.Array, axis: int, reverse: bool = False
) -> jax.Array:
  """cumsum() implemented via matrix-multiplciation."""
  # This is much faster on TPUs than the default cumsum.
  if not -x.ndim <= axis < x.ndim:
    raise ValueError(f'invalid {axis=}')
  if axis < 0:
    axis = axis + x.ndim
  size = x.shape[axis]
  i = jnp.arange(size)[:, jnp.newaxis]
  j = jnp.arange(size)[jnp.newaxis, :]
  op = jnp.greater_equal if reverse else jnp.less_equal
  w = op(i, j).astype(np.float32)
  out_axes = list(range(x.ndim))
  out_axes[axis] = x.ndim
  return jnp.einsum(
      w,
      [axis, x.ndim],
      x,
      list(range(x.ndim)),
      out_axes,
      precision=('bfloat16', 'highest'),
  )


def _parallel_dot_cumsum(
    x: jax.Array, axis: int, reverse: bool, axis_name: str
) -> jax.Array:
  """Parallel implementation of dot cumsum using lax primitives."""
  partials = _single_device_dot_cumsum(x, axis=axis, reverse=reverse)
  last_partial = lax.index_in_dim(partials, 0 if reverse else -1, axis)
  sums = lax.all_gather(last_partial, axis_name, tiled=True)
  axis_index = lax.axis_index(axis_name)
  op = jnp.greater if reverse else jnp.less
  total = partials
  terms = sums[1:] if reverse else sums[:-1]
  start = 1 if reverse else 0
  for i, term in enumerate(terms, start=start):
    total += op(i, axis_index) * term
  return total


def _dot_cumsum(
    x: jax.Array,
    axis: int,
    sharding: jax.sharding.NamedSharding | None,
    reverse: bool = False,
) -> jax.Array:
  """Dot product based cumsum() with custom parallel implementation."""
  if sharding is None or sharding.spec[axis] is None:
    # axis being summed along is not sharded
    return _single_device_dot_cumsum(x, axis, reverse=reverse)

  mesh = sharding.mesh
  spec = sharding.spec

  @jax.jit
  @functools.partial(
      shard_map.shard_map,
      mesh=mesh,
      in_specs=(spec,),
      out_specs=spec,
      check_rep=False,
  )
  def dot_cumsum(x):
    return _parallel_dot_cumsum(
        x, axis=axis, reverse=reverse, axis_name=sharding.spec[axis]
    )

  return dot_cumsum(x)


def cumsum(
    x: np.ndarray | jax.Array,
    axis: int,
    method: str = 'dot',
    sharding: jax.sharding.NamedSharding | None = None,
) -> jax.Array:
  """Optimized version of cumsum for short axes on TPUs."""
  if method == 'dot':
    return _dot_cumsum(x, axis, sharding=sharding)
  elif method == 'jax':
    return jnp.cumsum(x, axis)
  else:
    raise ValueError(f'invalid {method=}')


def reverse_cumsum(
    x: np.ndarray | jax.Array,
    axis: int,
    method: str = 'dot',
    sharding: jax.sharding.NamedSharding | None = None,
) -> jax.Array:
  """Performs cumsum in reverse order along `axis`."""
  if method == 'dot':
    return _dot_cumsum(x, axis, reverse=True, sharding=sharding)
  elif method == 'jax':
    return jnp.flip(jnp.cumsum(jnp.flip(x, axis), axis), axis)
  else:
    raise ValueError(f'invalid {method=}')


# XLA can generate rather inefficient code for Gather and DynamicSlice,
# especially on TPUs and when using automatic parallelization (because Gather
# can in principle be between arbitrary machines), batching (because the
# batching rule for DynamicSlice turns it into Gather) or reverse mode automatic
# differentiation (because the transpose of Gather is Scatter, which TPUs
# implement even less efficiently than Gather).
#
# To work around this issue, we need to reimplement jax.numpy operations
# (indexing regular array indexing) in terms of lax.slice and lax.pad, which are
# much easier for XLA to optimize.


def pad_in_dim(
    x: np.ndarray | jax.Array, pad_width: tuple[int, int], axis: int
) -> jax.Array:
  """Pad an array along an axis."""
  padding_value = jnp.array(0, dtype=x.dtype)
  padding_config = [(0, 0, 0)] * x.ndim
  padding_config[axis] = pad_width + (0,)  # add "interior" padding for lax.pad
  return lax.pad(x, padding_value, padding_config)


def shift(x: np.ndarray | jax.Array, offset: int, axis: int) -> jax.Array:
  """Like jnp.roll, but zero pads and implemented with different primitives."""
  if abs(offset) >= x.shape[axis]:
    return jnp.zeros_like(x)
  if offset > 0:
    sliced = lax.slice_in_dim(x, 0, x.shape[axis] - offset, axis=axis)
    return pad_in_dim(sliced, (offset, 0), axis=axis)
  else:
    sliced = lax.slice_in_dim(x, -offset, x.shape[axis], axis=axis)
    return pad_in_dim(sliced, (0, -offset), axis=axis)


def diff(x, axis=-1):
  """Like jnp.diff, but implemented in terms of simpler primitives."""
  upper = lax.slice_in_dim(x, 1, None, axis=axis)
  lower = lax.slice_in_dim(x, 0, -1, axis=axis)
  return upper - lower


def _reversed_arg_order_einsum(
    subscripts: str, x: jax.Array, y: jax.Array, **kwargs: ...
) -> jax.Array:
  # Surprisingly, XLA can generate code with up to ~10% speed differences
  # depending on the order of einsum arguments. There is no consistent pattern
  # about which is better.
  in_subscripts, out_subscripts = subscripts.split('->')
  lhs_subscripts, rhs_subscripts = in_subscripts.split(',')
  new_subscripts = f'{rhs_subscripts},{lhs_subscripts}->{out_subscripts}'
  return jnp.einsum(new_subscripts, y, x, **kwargs)


def _allgather_matmul_twoway(
    einsum_spec: str,
    lhs: jax.Array,
    rhs: jax.Array,
    split_axis: int,
    axis_name: str | tuple[str, str],
    reverse_arg_order: bool = False,
    precision: str = 'float32',
) -> jax.Array:
  """All-gather matmul using two way communication.

  This function is designed to be called from within `shmap`. It transfers
  inputs chunks using two way communication along a JAX mesh dimension, and then
  computes matrix-multiplication on each chunk.

  Args:
    einsum_spec: matmul specification.
    lhs: constant coefficients.
    rhs: sharded inputs.
    split_axis: the axis of `lhs` to split.
    axis_name: name of the axis to reduce along.
    reverse_arg_order: whether to reverse the order of arguments when calling
      einsum on chunks.
    precision: floating point precision to use for einsum.

  Returns:
    Result of einsum operation.
  """
  # adapted/simplified from
  # http://google3/third_party/google_research/google_research/scaling_transformer_inference_efficiency/collectives.py;l=327;rcl=521587778
  axis_size = lax.psum(1, axis_name)

  matmul = functools.partial(
      _reversed_arg_order_einsum if reverse_arg_order else jnp.einsum,
      einsum_spec,
      precision=precision,
  )

  if axis_size == 1:
    return matmul(lhs, rhs)

  if axis_size % 2:
    raise ValueError(f'axis_size must be 1 or even: {axis_size}')

  axis_index = lax.axis_index(axis_name)
  chunk_size = lhs.shape[split_axis] // axis_size

  def get_lhs_chunk(i):
    chunk_index = (axis_index + i) % axis_size
    lhs_chunk = lax.dynamic_slice_in_dim(
        lhs, chunk_index * chunk_size, chunk_size, axis=split_axis
    )
    return lhs_chunk

  def indexed_computation(i, rhs_fwd, rhs_bwd):
    lhs_fwd = get_lhs_chunk(-i)
    lhs_bwd = get_lhs_chunk(i + 1)
    return matmul(lhs_fwd, rhs_fwd) + matmul(lhs_bwd, rhs_bwd)

  perm_fwd = [(j, (j + 1) % axis_size) for j in range(axis_size)]
  perm_bwd = [(j, (j - 1) % axis_size) for j in range(axis_size)]

  def collective_matmul(i, carrys):
    accum, rhs_fwd, rhs_bwd = carrys
    rhs_fwd = lax.ppermute(rhs_fwd, axis_name, perm=perm_fwd)
    rhs_bwd = lax.ppermute(rhs_bwd, axis_name, perm=perm_bwd)
    accum += indexed_computation(i, rhs_fwd, rhs_bwd)
    return accum, rhs_fwd, rhs_bwd

  rhs_fwd = rhs
  rhs_bwd = lax.ppermute(rhs, axis_name, perm=perm_bwd)
  accum = indexed_computation(0, rhs_fwd, rhs_bwd)

  accum, rhs_fwd, rhs_bwd = lax.fori_loop(
      1, axis_size // 2, collective_matmul, (accum, rhs_fwd, rhs_bwd)
  )
  return accum


def _matmul_reducescatter_twoway(
    einsum_spec: str,
    lhs: jax.Array,
    rhs: jax.Array,
    scatter_axis: int,
    axis_name: str | tuple[str, str],
    reverse_arg_order: bool = False,
    precision: str = 'float32',
) -> jax.Array:
  """All-gather matmul using two way communication.

  This function is designed to be called from within `shmap`. It computes
  matrix-multiplication on each chunk, and then transfers the result using two
  way communication along a JAX mesh dimension.

  Args:
    einsum_spec: matmul specification.
    lhs: constant coefficients.
    rhs: sharded inputs.
    scatter_axis: the axis of `lhs` to scatter along.
    axis_name: name of the axis to reduce along.
    reverse_arg_order: whether to reverse the order of arguments when calling
      einsum on chunks.
    precision: floating point precision to use for einsum.

  Returns:
    Result of einsum operation.
  """
  # adapted/simplified from
  # http://google3/third_party/google_research/google_research/scaling_transformer_inference_efficiency/collectives.py;l=671;rcl=521587778
  axis_size = lax.psum(1, axis_name)
  matmul = functools.partial(
      _reversed_arg_order_einsum if reverse_arg_order else jnp.einsum,
      einsum_spec,
      precision=precision,
  )

  if axis_size == 1:
    return matmul(lhs, rhs)

  if axis_size % 2:
    raise ValueError(f'axis_size must be 1 or even: {axis_size}')

  axis_index = lax.axis_index(axis_name)
  chunk_size = lhs.shape[scatter_axis] // axis_size

  def indexed_computation(i):
    chunk_index = (axis_index + axis_size // 2 + i) % axis_size
    lhs_chunk = lax.dynamic_slice_in_dim(
        lhs, chunk_index * chunk_size, chunk_size, axis=scatter_axis
    )
    return matmul(lhs_chunk, rhs)

  perm_fwd = [(j, (j + 1) % axis_size) for j in range(axis_size)]
  perm_bwd = [(j, (j - 1) % axis_size) for j in range(axis_size)]

  def collective_matmul(i, carrays):
    accum_fwd, accum_bwd = carrays
    accum_fwd = lax.ppermute(accum_fwd, axis_name, perm=perm_fwd)
    accum_bwd = lax.ppermute(accum_bwd, axis_name, perm=perm_bwd)
    accum_fwd += indexed_computation(-i)
    accum_bwd += indexed_computation(i + 1)
    return accum_fwd, accum_bwd

  accum_fwd = indexed_computation(0)
  accum_bwd = indexed_computation(1)

  (accum_fwd, accum_bwd) = lax.fori_loop(
      1, axis_size // 2, collective_matmul, (accum_fwd, accum_bwd)
  )
  accum_fwd = lax.ppermute(accum_fwd, axis_name, perm=perm_fwd)
  accum = accum_fwd + accum_bwd
  return accum


def _parse_einsum_subscripts(subscripts: str) -> tuple[str, str, str]:
  if '...' in subscripts:
    raise ValueError(f'ellipsis not supported by sharded_einsum: {subscripts}')
  matches = re.fullmatch(r'(\w+)\,(\w+)\-\>(\w+)', subscripts)
  if matches is None:
    raise ValueError(f'{subscripts=} does not match the expected pattern')
  lhs_subscripts, rhs_subscripts, out_subscripts = matches.groups()
  return lhs_subscripts, rhs_subscripts, out_subscripts


def _determine_reduce_subscript(
    lhs_subscripts: str,
    rhs_subscripts: str,
    out_subscripts: str,
    rhs_spec: jax.sharding.PartitionSpec,
) -> str:
  """Determine the reduce axis subscript from einsum subscripts."""
  subscripts = []
  for subscript in lhs_subscripts:
    if (
        subscript not in out_subscripts
        and subscript in rhs_subscripts
        and rhs_spec[rhs_subscripts.index(subscript)] is not None
    ):
      subscripts.append(subscript)
  if len(subscripts) != 1:
    raise ValueError(f'multiple sharded axes are reduced over: {subscripts}')
  (subscript,) = subscripts  # pylint: disable=unbalanced-tuple-unpacking
  return subscript


def _determine_transfer_subscript(
    lhs_subscripts: str,
    rhs_subscripts: str,
    out_subscripts: str,
    out_spec: jax.sharding.PartitionSpec,
) -> str:
  """Determine the transfer axis subscript from einsum subscripts."""
  subscripts = []
  for subscript in lhs_subscripts:
    if (
        subscript not in rhs_subscripts
        and subscript in out_subscripts
        and out_spec[out_subscripts.index(subscript)] is not None
    ):
      subscripts.append(subscript)
  if len(subscripts) != 1:
    raise ValueError(
        f'multiple sharded axes are transferred over: {subscripts}'
    )
  (subscript,) = subscripts  # pylint: disable=unbalanced-tuple-unpacking
  return subscript


@jax.named_call
def sharded_einsum(
    subscripts: str,
    lhs: np.ndarray | jax.Array,
    rhs: jax.Array,
    /,
    *,
    gather_inputs: bool | None = None,
    reverse_arg_order: bool = False,
    precision: str = 'float32',
    mesh: jax.sharding.Mesh | None,
    rhs_spec: jax.sharding.PartitionSpec,
    out_spec: jax.sharding.PartitionSpec,
) -> jax.Array:
  """Calculate a two argument einsum with a single sharded inputs.

  Designed for supporting `einsum` patterns that arise in the Neural GCM dycore,
  i.e., forward/inverse Fourier and Legendre transforms. Only reductions along
  a single sharded axis are supported.

  Args:
    subscripts: einsum specification string.
    lhs: non-sharded left hand side argument, in the form of a static NumPy or
      JAX array.
    rhs: right hand side argument, potentially sharded along any axes.
    gather_inputs: if `True`, transfer input data between cores to implement the
      parallel einsum; if `False`, transfer output data from the result of
      calculating each chunk. The default behavior is to use whichever strategy
      results in less data movement, i.e., `gather_inputs=False` iff the outputs
      are smaller than the `rhs``.
    reverse_arg_order: if True, call einsum on each chunk like `einsum(..., rhs,
      lhs)` instead of `einsum(..., lhs, rhs)`. This results in different XLA
      optimizations and can occasionally be somewhat faster.
    precision: floating point precision to use for einsum.
    mesh: parallel mesh to use for implementing this operation, or `None`, which
      indicates no sharding.
    rhs_spec: sharding spec for `rhs`.
    out_spec: sharding spec for outputs. Dimensions in `rhs` that are preserved
      in the outputs should be chunked in the way on `rhs_spec`.

  Returns:
    Result of `einsum(subscripts, lhs, rhs)`, sharded according to `out_spec`.
  """
  all_subscripts = _parse_einsum_subscripts(subscripts)
  lhs_subscripts, rhs_subscripts, out_subscripts = all_subscripts

  if mesh is None:
    # validation complete; calculate non-sharded einsum if no mesh supplied
    return jnp.einsum(subscripts, lhs, rhs, precision=precision)

  reduce_subscript = _determine_reduce_subscript(*all_subscripts, rhs_spec)
  transfer_subscript = _determine_transfer_subscript(*all_subscripts, out_spec)
  reduce_axis_name = rhs_spec[rhs_subscripts.index(reduce_subscript)]

  if gather_inputs is None:
    out_shape = jax.eval_shape(
        functools.partial(jnp.einsum, subscripts), lhs, rhs
    ).shape
    gather_inputs = math.prod(out_shape) > math.prod(rhs.shape)

  if gather_inputs:
    lhs_partitions = [
        out_spec[out_subscripts.index(i)] if i in out_subscripts else None
        for i in lhs_subscripts
    ]
    split_axis = lhs_subscripts.index(reduce_subscript)
    matmul_impl = functools.partial(
        _allgather_matmul_twoway, split_axis=split_axis
    )
  else:
    lhs_partitions = [
        rhs_spec[rhs_subscripts.index(i)] if i in rhs_subscripts else None
        for i in lhs_subscripts
    ]
    scatter_axis = lhs_subscripts.index(transfer_subscript)
    matmul_impl = functools.partial(
        _matmul_reducescatter_twoway, scatter_axis=scatter_axis
    )

  lhs_spec = jax.sharding.PartitionSpec(*lhs_partitions)
  in_specs = (lhs_spec, rhs_spec)

  @functools.partial(
      shard_map.shard_map,
      mesh=mesh,
      in_specs=in_specs,
      out_specs=out_spec,
      check_rep=False,  # https://github.com/google/jax/issues/15894
  )
  @jax.jit  # https://github.com/google/jax/issues/15723
  def distributed_matmul(lhs, rhs):
    return matmul_impl(
        subscripts,
        lhs,
        rhs,
        axis_name=reduce_axis_name,
        precision=precision,
        reverse_arg_order=reverse_arg_order,
    )

  return distributed_matmul(lhs, rhs)



# Copyright 2023 Google LLC

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     https://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from absl.testing import absltest
from absl.testing import parameterized
from dinosaur import sigma_coordinates
from dinosaur import spherical_harmonic
import numpy as np


def _broadcast(*args):
  """Reshapes `args` so that they will broadcast over `len(args)` dimensions."""
  broadcasted = []
  for j, arg in enumerate(args):
    shape = [1] * len(args)
    shape[j] = -1
    broadcasted.append(arg.reshape(shape))
  return broadcasted


# pylint: disable=unbalanced-tuple-unpacking


def quadratic_function(sigma, lon, lat):
  """A test function for vertical differentiation and integration."""
  sigma, lon, lat = _broadcast(sigma, lon, lat)
  return sigma**2 * (1 + np.cos(lon) * np.cos(lat))


def quadratic_derivative(sigma, lon, lat):
  """The derivative of `quadratic_function` with respect to `sigma`."""
  sigma, lon, lat = _broadcast(sigma, lon, lat)
  return 2 * sigma * (1 + np.cos(lon) * np.cos(lat))


def quadratic_integral(sigma, lon, lat):
  """The indefinite integral of `quadratic_function` with respect to `sigma`."""
  sigma, lon, lat = _broadcast(sigma, lon, lat)
  return sigma**3 / 3 * (1 + np.cos(lon) * np.cos(lat))


def exponential_function(sigma, lon, lat):
  """A test function for vertical differentiation and integration."""
  sigma, lon, lat = _broadcast(sigma, lon, lat)
  return np.exp(sigma) * np.cos(lon) * np.sin(lat)


exponential_derivative = exponential_integral = exponential_function

# pylint: enable=unbalanced-tuple-unpacking


def test_cases():
  # Note that the test cases use far more layers that we will use in practice.
  # We do this to test numerical accuracy against closed form derivatives and
  # integrals.
  return (
      dict(
          testcase_name='Quadratic',
          test_function=quadratic_function,
          derivative_function=quadratic_derivative,
          integral_function=quadratic_integral,
          layers=np.array([10, 20, 40, 80, 160, 320]),
          grid_resolution=8,
      ),
      dict(
          testcase_name='Exponential',
          test_function=exponential_function,
          derivative_function=exponential_derivative,
          integral_function=exponential_integral,
          layers=np.array([10, 20, 40, 80, 160]),
          grid_resolution=16,
      ),
  )


def _test_error_scaling(layers, errors, error_scaling):
  """Checks that `errors` scales with `layers` according to `error_scaling`."""
  log_error_ratios = np.diff(np.log(errors))
  log_expected_ratios = np.diff(np.log(error_scaling(layers)))
  np.testing.assert_allclose(log_error_ratios, log_expected_ratios, atol=0.05)


class SigmaCoordinatesTest(parameterized.TestCase):

  @parameterized.parameters(
      ([0.0, 0.5, 1.0],),
      ((0.0, 0.5, 1.0),),
      (np.array([0.0, 0.5, 1.0]),),
  )
  def test_initialization_casting(self, boundaries):
    coordinates = sigma_coordinates.SigmaCoordinates(boundaries)
    self.assertIsInstance(coordinates.boundaries, np.ndarray)

    with self.subTest('asdict'):
      self.assertIsInstance(coordinates.asdict()['boundaries'], list)
      np.testing.assert_array_equal(
          coordinates.asdict()['boundaries'], boundaries
      )

  def test_initialization_raises(self):
    with self.subTest('end values'):
      with self.assertRaisesWithLiteralMatch(
          ValueError,
          'Expected boundaries[0] = 0, boundaries[-1] = 1, '
          'got boundaries = [0.2 0.5 1. ]',
      ):
        sigma_coordinates.SigmaCoordinates([0.2, 0.5, 1])
      with self.assertRaisesWithLiteralMatch(
          ValueError,
          'Expected boundaries[0] = 0, boundaries[-1] = 1, '
          'got boundaries = [0.  0.5 0.9]',
      ):
        sigma_coordinates.SigmaCoordinates([0.0, 0.5, 0.9])
    with self.subTest('increasing'):
      with self.assertRaisesWithLiteralMatch(
          ValueError,
          'Expected `boundaries` to be monotonically increasing, '
          'got boundaries = [0.  0.5 0.5 1. ]',
      ):
        sigma_coordinates.SigmaCoordinates([0.0, 0.5, 0.5, 1])

  def test_from_centers(self):
    centers = np.array([0.1, 0.35, 0.75], dtype=np.float32)
    coordinates = sigma_coordinates.SigmaCoordinates.from_centers(centers)
    expected_boundaries = np.array([0.0, 0.2, 0.5, 1.0], dtype=np.float32)
    np.testing.assert_array_equal(coordinates.boundaries, expected_boundaries)
    np.testing.assert_array_equal(coordinates.centers, centers)

  @parameterized.named_parameters(*test_cases())
  def test_centered_difference(
      self, test_function, derivative_function, layers, grid_resolution, **_
  ):
    """Tests `centered_difference` against the closed form derivative."""
    grid = spherical_harmonic.Grid.with_wavenumbers(grid_resolution)
    lon, lat = grid.nodal_axes
    # Note that we only test accuracy for derivatives at the finest resolution.
    coordinates = sigma_coordinates.SigmaCoordinates.equidistant(layers[-1])
    centers = coordinates.centers
    boundaries = coordinates.internal_boundaries
    x = test_function(centers, lon, lat)
    expected_derivative = derivative_function(boundaries, lon, lat)
    computed_derivative = sigma_coordinates.centered_difference(x, coordinates)
    np.testing.assert_allclose(
        expected_derivative, computed_derivative, atol=1e-3
    )

  @parameterized.named_parameters(*test_cases())
  def test_cumulative_sigma_integral_downward(
      self, test_function, integral_function, layers, grid_resolution, **_
  ):
    """Tests `sigma_integral` in the downward direction."""
    grid = spherical_harmonic.Grid.with_wavenumbers(grid_resolution)
    lon, lat = grid.nodal_axes
    total_errors = []
    for nlayers in layers:
      coordinates = sigma_coordinates.SigmaCoordinates.equidistant(nlayers)
      centers = coordinates.centers
      boundaries = coordinates.boundaries
      x = test_function(centers, lon, lat)
      indefinite_integral = integral_function(boundaries, lon, lat)
      expected_integral = indefinite_integral[1:] - indefinite_integral[0]
      computed_integral = sigma_coordinates.cumulative_sigma_integral(
          x, coordinates, cumsum_method='jax'
      )
      computed_integral_all = sigma_coordinates.sigma_integral(
          x, coordinates, keepdims=False
      )
      np.testing.assert_allclose(
          computed_integral[-1], computed_integral_all, atol=1e-6
      )
      # To test convergence, we compute the error in the integral at the
      # "bottom" layer.
      total_errors.append(
          np.abs(expected_integral[-1] - computed_integral[-1]).max()
      )
    with self.subTest('Convergence'):
      # Since we use a midpoint method, we expect errors to scale with
      # 1 / layers².
      error_scaling = lambda l: 1 / l**2
      _test_error_scaling(layers, total_errors, error_scaling)
    with self.subTest('Accuracy'):
      # Tests that the integral is accurate at the highest resolution.
      np.testing.assert_allclose(
          expected_integral[-1], computed_integral[-1], atol=1e-2
      )

  @parameterized.named_parameters(*test_cases())
  def test_cumulative_sigma_integral_upward(
      self, test_function, integral_function, layers, grid_resolution, **_
  ):
    """Tests `sigma_integral` in the upward direction."""
    grid = spherical_harmonic.Grid.with_wavenumbers(grid_resolution)
    lon, lat = grid.nodal_axes
    total_errors = []
    for nlayers in layers:
      coordinates = sigma_coordinates.SigmaCoordinates.equidistant(nlayers)
      centers = coordinates.centers
      boundaries = coordinates.boundaries
      x = test_function(centers, lon, lat)
      indefinite_integral = integral_function(boundaries, lon, lat)
      expected_integral = -(indefinite_integral[:-1] - indefinite_integral[-1])
      computed_integral = sigma_coordinates.cumulative_sigma_integral(
          x, coordinates, downward=False, cumsum_method='jax'
      )
      computed_integral_all = sigma_coordinates.sigma_integral(
          x, coordinates, keepdims=False
      )
      np.testing.assert_allclose(
          computed_integral[0], computed_integral_all, atol=1e-6
      )
      # To test convergence, we compute the error in the integral at the
      # "top" layer.
      total_errors.append(
          np.abs(expected_integral[0] - computed_integral[0]).max()
      )
    with self.subTest('Convergence'):
      # Since we use a midpoint method in  we expect errors to scale with
      # 1 / layers².
      error_scaling = lambda l: 1 / l**2
      _test_error_scaling(layers, total_errors, error_scaling)
    with self.subTest('Accuracy'):
      # Tests that the integral is accurate at the highest resolution.
      np.testing.assert_allclose(
          expected_integral[0], computed_integral[0], atol=1e-2
      )

  @parameterized.named_parameters(*test_cases())
  def test_log_sigma_integral_downward(
      self, test_function, integral_function, layers, grid_resolution, **_
  ):
    """Tests `cumulative_log_sigma_integral` in the downward direction."""
    grid = spherical_harmonic.Grid.with_wavenumbers(grid_resolution)
    lon, lat = grid.nodal_axes
    total_errors = []
    for nlayers in layers:
      coordinates = sigma_coordinates.SigmaCoordinates.equidistant(nlayers)
      centers = coordinates.centers
      broadcasted_centers = centers[:, np.newaxis, np.newaxis]
      # We integrate ∫f(x) 𝜎 d(log𝜎) =  ∫f(x) d𝜎
      x = test_function(centers, lon, lat) * broadcasted_centers
      indefinite_integral = integral_function(centers, lon, lat)
      integral_boundary = integral_function(np.zeros(1), lon, lat)
      expected_integral = indefinite_integral - integral_boundary
      computed_integral = sigma_coordinates.cumulative_log_sigma_integral(
          x, coordinates
      )
      # To test convergence, we compute the error in the integral at the
      # "bottom" layer.
      total_errors.append(
          np.abs(expected_integral[-1] - computed_integral[-1]).max()
      )
    with self.subTest('Convergence'):
      # Since we use a trapezoidal method in log space, we expect errors to
      # scale with the inverse square of the spacing in log space. Note that if
      # the boundaries were equidistant in log space, this would be equivalent
      # to the usual 1 / layers² scaling.
      def error_scaling(layers):
        expected_scaling = []
        for l in layers:
          centers = sigma_coordinates.SigmaCoordinates.equidistant(l).centers
          log_space_widths = np.diff(np.log(centers))
          expected_scaling.append(np.square(log_space_widths).mean())
        return np.array(expected_scaling)

      _test_error_scaling(layers, total_errors, error_scaling)
    with self.subTest('Accuracy'):
      # Tests that the integral is accurate at the highest resolution.
      np.testing.assert_allclose(
          expected_integral[-1], computed_integral[-1], atol=1e-2
      )

  @parameterized.named_parameters(*test_cases())
  def test_log_sigma_integral_upward(
      self, test_function, integral_function, layers, grid_resolution, **_
  ):
    """Tests `cumulative_log_sigma_integral` in the upward direction."""
    grid = spherical_harmonic.Grid.with_wavenumbers(grid_resolution)
    lon, lat = grid.nodal_axes
    total_errors = []
    for nlayers in layers:
      coordinates = sigma_coordinates.SigmaCoordinates.equidistant(nlayers)
      centers = coordinates.centers
      broadcasted_centers = centers[:, np.newaxis, np.newaxis]
      # We integrate ∫f(x) 𝜎 d(log𝜎) =  ∫f(x) d𝜎
      x = test_function(centers, lon, lat) * broadcasted_centers
      indefinite_integral = integral_function(centers, lon, lat)
      integral_boundary = integral_function(
          np.array([coordinates.centers[-1]]), lon, lat
      )
      expected_integral = -(indefinite_integral - integral_boundary)
      computed_integral = sigma_coordinates.cumulative_log_sigma_integral(
          x, coordinates, downward=False
      )
      # To test convergence, we compute the error in the integral at the
      # "bottom" layer.
      total_errors.append(
          np.abs(expected_integral[0] - computed_integral[0]).max()
      )
    print(f'TOTAL ERROR: {total_errors}')
    with self.subTest('Convergence'):
      # Since we use a trapezoidal method in log space, we expect errors to
      # scale with the inverse square of the spacing in log space. Note that if
      # the boundaries were equidistant in log space, this would be equivalent
      # to the usual 1 / layers² scaling.
      def error_scaling(layers):
        expected_scaling = []
        for l in layers:
          centers = sigma_coordinates.SigmaCoordinates.equidistant(l).centers
          log_space_widths = np.diff(np.log(centers))
          expected_scaling.append(np.square(log_space_widths).mean())
        return np.array(expected_scaling)

      _test_error_scaling(layers, total_errors, error_scaling)
    with self.subTest('Accuracy'):
      # Tests that the integral is accurate at the highest resolution.
      np.testing.assert_allclose(
          expected_integral[0], computed_integral[0], atol=1e-2
      )

  @parameterized.named_parameters(*test_cases())
  def test_vertical_advection_helper(
      self, test_function, derivative_function, layers, grid_resolution, **_
  ):
    """Tests `sigma_coordinates.centered_vertical_advection` helper function."""
    grid = spherical_harmonic.Grid.with_wavenumbers(grid_resolution)
    lon, lat = grid.nodal_axes
    # Note that we only test accuracy for derivatives at the finest resolution.
    coordinates = sigma_coordinates.SigmaCoordinates.equidistant(layers[-1])
    centers = coordinates.centers
    boundaries = coordinates.internal_boundaries
    velocity_fn = test_function
    derivative_fn = derivative_function

    x = test_function(centers, lon, lat)
    dx_dsigma = derivative_fn(centers, lon, lat)
    w = velocity_fn(boundaries, lon, lat)

    with self.subTest('Default boundary conditions, centered'):
      # By default boundary conditions are set to zeros.
      w_boundary_values = np.zeros_like(w[:2, ...])
      dx_dsigma_boundary_values = np.zeros_like(dx_dsigma[:2, ...])
      # This does not affect the bulk, but modifies boundary to:
      # -(w[inner] * ∂x/∂𝜎[inner] + w[ghost] * ∂x/∂𝜎[ghost]) / 2
      boundaries = -0.5 * (
          w[[0, -1], ...] * derivative_fn(boundaries, lon, lat)[[0, -1], ...]
          + w_boundary_values * dx_dsigma_boundary_values
      )
      expected = -dx_dsigma * velocity_fn(centers, lon, lat)
      expected[[0, -1], ...] = boundaries
      actual = sigma_coordinates.centered_vertical_advection(w, x, coordinates)
      np.testing.assert_allclose(actual, expected, atol=1e-3)

    with self.subTest('Default boundary conditions, upwind'):
      actual = sigma_coordinates.upwind_vertical_advection(w, x, coordinates)
      # Upwinding is only 1st order accurate (vs 2nd order centered advection)
      np.testing.assert_allclose(actual[1:-1], expected[1:-1], atol=5e-2)

    with self.subTest('Custom boundary conditions'):
      # Tests that boundary values can be provided.
      w_boundary_values = velocity_fn(coordinates.boundaries[[0, -1]], lon, lat)
      w_boundary_values = (
          w_boundary_values[[0], ...],
          w_boundary_values[[1], ...],
      )
      dx_dsigma_boundary_values = derivative_fn(
          coordinates.boundaries[[0, -1]], lon, lat
      )
      dx_dsigma_boundary_values = (
          dx_dsigma_boundary_values[[0], ...],
          dx_dsigma_boundary_values[[1], ...],
      )
      # with these values boundaries should remain unmodified.
      expected = -dx_dsigma * velocity_fn(centers, lon, lat)
      actual = sigma_coordinates.centered_vertical_advection(
          w,
          x,
          coordinates,
          w_boundary_values=w_boundary_values,
          dx_dsigma_boundary_values=dx_dsigma_boundary_values,
      )
      np.testing.assert_allclose(actual, expected, atol=1e-3)


if __name__ == '__main__':
  absltest.main()



# Copyright 2023 Google LLC

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     https://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import functools
from typing import Any

from absl.testing import absltest
from absl.testing import parameterized
from dinosaur import coordinate_systems
from dinosaur import horizontal_interpolation
from dinosaur import layer_coordinates
from dinosaur import primitive_equations
from dinosaur import primitive_equations_states
from dinosaur import scales
from dinosaur import shallow_water
from dinosaur import shallow_water_states
from dinosaur import sigma_coordinates
from dinosaur import spherical_harmonic
from dinosaur import time_integration
from dinosaur import vertical_interpolation
from dinosaur import xarray_utils
import jax
import numpy as np
import pandas as pd
import xarray


def seed_stream(s):
  k = jax.random.PRNGKey(s)
  while True:
    k, l = jax.random.split(k)
    yield l


def shape_structure(inputs):
  return jax.tree.map(lambda x: x.shape, inputs)


class XarrayUtilsTest(parameterized.TestCase):

  def _check_attrs(self, dataset, grid_attrs, custom_attrs):
    """Helper function that asserts that dataset has expected attrs."""
    for attr in grid_attrs:
      self.assertTrue(hasattr(dataset, attr))
    if custom_attrs is not None:
      for attr in custom_attrs.keys():
        self.assertTrue(hasattr(dataset, attr))

  def assert_values_shape_equal(self, x: Any, y: Any):
    if not isinstance(x, dict):
      x = x.asdict()
    if not isinstance(y, dict):
      y = y.asdict()
    self.assertDictEqual(shape_structure(x), shape_structure(y))

  @parameterized.parameters(
      dict(
          samples=2,
          time_steps=3,
          dt=0.6,
          layers=8,
          wavenumbers=32,
          attrs=dict(g=9.80616),
      ),
      dict(
          samples=4,
          time_steps=2,
          dt=300.0,
          layers=2,
          wavenumbers=32,
          attrs=dict(a=9.80616, mean=10.5),
      ),
  )
  def test_primitive_eq_data_to_xarray(
      self, samples, time_steps, dt, layers, wavenumbers, attrs
  ):
    physics_specs = primitive_equations.PrimitiveEquationsSpecs.from_si()
    grid = spherical_harmonic.Grid.with_wavenumbers(
        wavenumbers, radius=physics_specs.radius
    )
    vertical_grid = sigma_coordinates.SigmaCoordinates.equidistant(layers)
    coords = coordinate_systems.CoordinateSystem(grid, vertical_grid)

    # creating state, trajectory and batch of trajectories to be converted.
    initial_state_fn, _ = primitive_equations_states.steady_state_jw(
        coords, physics_specs
    )
    state = initial_state_fn()
    state.tracers = {
        'test_tracer': primitive_equations_states.gaussian_scalar(
            coords, physics_specs, amplitude=0.1
        )
    }
    trajectory = jax.tree.map(
        lambda *args: np.stack(args), *([state] * time_steps)
    )
    batch_of_trajectories = jax.tree.map(
        lambda *args: np.stack(args), *([trajectory] * samples)
    )
    expected_grid_attrs = coords.asdict()

    with self.subTest('state_to_dataset_modal'):
      expected_coords_sizes = {
          xarray_utils.XR_LON_MODE_NAME: grid.longitude_wavenumbers * 2 - 1,
          xarray_utils.XR_LAT_MODE_NAME: grid.total_wavenumbers,
          xarray_utils.XR_LEVEL_NAME: vertical_grid.layers,
          xarray_utils.XR_SURFACE_NAME: 1,
      }
      ds = xarray_utils.data_to_xarray(
          state.asdict(),
          sample_ids=None,
          times=None,
          coords=coords,
          attrs=attrs,
      )
      self._check_attrs(ds, expected_grid_attrs, attrs)
      self.assertDictEqual(dict(ds.sizes), expected_coords_sizes)
      reconstruction = xarray_utils.xarray_to_primitive_eq_data(
          ds, tracers_to_include=('test_tracer',)
      )
      self.assert_values_shape_equal(state, reconstruction)

    with self.subTest('trajectory_to_dataset_modal'):
      expected_coords_sizes = {
          xarray_utils.XR_LON_MODE_NAME: grid.longitude_wavenumbers * 2 - 1,
          xarray_utils.XR_LAT_MODE_NAME: grid.total_wavenumbers,
          xarray_utils.XR_LEVEL_NAME: vertical_grid.layers,
          xarray_utils.XR_SURFACE_NAME: 1,
          xarray_utils.XR_TIME_NAME: time_steps,
      }
      times = dt * np.arange(time_steps)
      ds = xarray_utils.data_to_xarray(
          trajectory.asdict(),
          sample_ids=None,
          times=times,
          coords=coords,
          attrs=attrs,
      )
      self._check_attrs(ds, expected_grid_attrs, attrs)
      self.assertDictEqual(dict(ds.sizes), expected_coords_sizes)
      reconstruction = xarray_utils.xarray_to_primitive_eq_data(
          ds, tracers_to_include=('test_tracer',)
      )
      self.assert_values_shape_equal(trajectory, reconstruction)

    with self.subTest('trajectory_to_dataset_modal'):
      expected_coords_sizes = {
          xarray_utils.XR_LON_MODE_NAME: grid.longitude_wavenumbers * 2 - 1,
          xarray_utils.XR_LAT_MODE_NAME: grid.total_wavenumbers,
          xarray_utils.XR_LEVEL_NAME: vertical_grid.layers,
          xarray_utils.XR_SURFACE_NAME: 1,
          xarray_utils.XR_TIME_NAME: time_steps,
          xarray_utils.XR_SAMPLE_NAME: samples,
      }
      times = dt * np.arange(time_steps)
      sample_ids = np.arange(samples)
      ds = xarray_utils.data_to_xarray(
          batch_of_trajectories.asdict(),
          sample_ids=sample_ids,
          times=times,
          coords=coords,
          attrs=attrs,
      )
      self._check_attrs(ds, expected_grid_attrs, attrs)
      self.assertDictEqual(dict(ds.sizes), expected_coords_sizes)
      reconstruction = xarray_utils.xarray_to_primitive_eq_data(
          ds, tracers_to_include=('test_tracer',)
      )
      self.assert_values_shape_equal(batch_of_trajectories, reconstruction)

    with self.subTest('trajectory_to_dataset_nodal'):
      expected_coords_sizes = {
          xarray_utils.XR_LON_NAME: grid.longitude_nodes,
          xarray_utils.XR_LAT_NAME: grid.latitude_nodes,
          xarray_utils.XR_LEVEL_NAME: vertical_grid.layers,
          xarray_utils.XR_SURFACE_NAME: 1,
          xarray_utils.XR_TIME_NAME: time_steps,
          xarray_utils.XR_SAMPLE_NAME: samples,
      }
      times = dt * np.arange(time_steps)
      sample_ids = np.arange(samples)
      nodal_batch_of_trajectories = jax.tree.map(
          grid.to_nodal, batch_of_trajectories
      )
      ds = xarray_utils.data_to_xarray(
          nodal_batch_of_trajectories.asdict(),
          sample_ids=sample_ids,
          times=times,
          coords=coords,
          attrs=attrs,
      )
      self._check_attrs(ds, expected_grid_attrs, attrs)
      self.assertDictEqual(dict(ds.sizes), expected_coords_sizes)
      reconstruction = xarray_utils.xarray_to_primitive_eq_data(
          ds, tracers_to_include=('test_tracer',)
      )
      self.assert_values_shape_equal(
          nodal_batch_of_trajectories, reconstruction
      )

  @parameterized.parameters(
      dict(
          samples=2,
          time_steps=3,
          dt=0.6,
          layers=8,
          wavenumbers=32,
          attrs=dict(g=9.80616),
      ),
      dict(
          samples=4,
          time_steps=2,
          dt=300.0,
          layers=2,
          wavenumbers=32,
          attrs=dict(a=9.80616, mean=10.5),
      ),
  )
  def test_shallow_water_eq_data_to_xarray(
      self, samples, time_steps, dt, layers, wavenumbers, attrs
  ):
    physics_specs = shallow_water.ShallowWaterSpecs.from_si(
        np.ones((layers,)) * scales.WATER_DENSITY
    )
    grid = spherical_harmonic.Grid.with_wavenumbers(
        wavenumbers, radius=physics_specs.radius
    )
    vertical_grid = layer_coordinates.LayerCoordinates(layers)
    coords = coordinate_systems.CoordinateSystem(grid, vertical_grid)

    # creating state, trajectory and batch of trajectories to be converted.
    _, sin_lat = grid.nodal_mesh
    lat = np.arcsin(sin_lat)
    velocity_function = lambda lat: np.cos(lat) ** 2 / 5
    velocity = np.stack([velocity_function(lat)] * layers)
    state = shallow_water_states.multi_layer(
        velocity, physics_specs.densities, coords
    )

    trajectory = jax.tree.map(
        lambda *args: np.stack(args), *([state] * time_steps)
    )
    batch_of_trajectories = jax.tree.map(
        lambda *args: np.stack(args), *([trajectory] * samples)
    )
    expected_grid_attrs = coords.asdict()

    with self.subTest('state_to_dataset_modal'):
      expected_coords_sizes = {
          xarray_utils.XR_LON_MODE_NAME: grid.longitude_wavenumbers * 2 - 1,
          xarray_utils.XR_LAT_MODE_NAME: grid.total_wavenumbers,
          xarray_utils.XR_LEVEL_NAME: vertical_grid.layers,
      }
      ds = xarray_utils.data_to_xarray(
          state.asdict(),
          sample_ids=None,
          times=None,
          coords=coords,
          attrs=attrs,
      )
      self._check_attrs(ds, expected_grid_attrs, attrs)
      self.assertDictEqual(dict(ds.sizes), expected_coords_sizes)
      reconstruction = xarray_utils.xarray_to_shallow_water_eq_data(ds)
      self.assert_values_shape_equal(state, reconstruction)

    with self.subTest('trajectory_to_dataset_modal'):
      expected_coords_sizes = {
          xarray_utils.XR_LON_MODE_NAME: grid.longitude_wavenumbers * 2 - 1,
          xarray_utils.XR_LAT_MODE_NAME: grid.total_wavenumbers,
          xarray_utils.XR_LEVEL_NAME: vertical_grid.layers,
          xarray_utils.XR_TIME_NAME: time_steps,
      }
      times = dt * np.arange(time_steps)
      ds = xarray_utils.data_to_xarray(
          trajectory.asdict(),
          sample_ids=None,
          times=times,
          coords=coords,
          attrs=attrs,
      )
      self._check_attrs(ds, expected_grid_attrs, attrs)
      self.assertDictEqual(dict(ds.sizes), expected_coords_sizes)
      reconstruction = xarray_utils.xarray_to_shallow_water_eq_data(ds)
      self.assert_values_shape_equal(trajectory, reconstruction)

    with self.subTest('trajectory_to_dataset_modal'):
      expected_coords_sizes = {
          xarray_utils.XR_LON_MODE_NAME: grid.longitude_wavenumbers * 2 - 1,
          xarray_utils.XR_LAT_MODE_NAME: grid.total_wavenumbers,
          xarray_utils.XR_LEVEL_NAME: vertical_grid.layers,
          xarray_utils.XR_TIME_NAME: time_steps,
          xarray_utils.XR_SAMPLE_NAME: samples,
      }
      times = dt * np.arange(time_steps)
      sample_ids = np.arange(samples)
      ds = xarray_utils.data_to_xarray(
          batch_of_trajectories.asdict(),
          sample_ids=sample_ids,
          times=times,
          coords=coords,
          attrs=attrs,
      )
      self._check_attrs(ds, expected_grid_attrs, attrs)
      self.assertDictEqual(dict(ds.sizes), expected_coords_sizes)
      reconstruction = xarray_utils.xarray_to_shallow_water_eq_data(ds)
      self.assert_values_shape_equal(batch_of_trajectories, reconstruction)

    with self.subTest('trajectory_to_dataset_nodal'):
      expected_coords_sizes = {
          xarray_utils.XR_LON_NAME: grid.longitude_nodes,
          xarray_utils.XR_LAT_NAME: grid.latitude_nodes,
          xarray_utils.XR_LEVEL_NAME: vertical_grid.layers,
          xarray_utils.XR_TIME_NAME: time_steps,
          xarray_utils.XR_SAMPLE_NAME: samples,
      }
      times = dt * np.arange(time_steps)
      sample_ids = np.arange(samples)
      nodal_batch_of_trajectories = jax.tree.map(
          grid.to_nodal, batch_of_trajectories
      )
      ds = xarray_utils.data_to_xarray(
          nodal_batch_of_trajectories.asdict(),
          sample_ids=sample_ids,
          times=times,
          coords=coords,
          attrs=attrs,
      )
      self._check_attrs(ds, expected_grid_attrs, attrs)
      self.assertDictEqual(dict(ds.sizes), expected_coords_sizes)
      reconstruction = xarray_utils.xarray_to_shallow_water_eq_data(ds)
      self.assert_values_shape_equal(
          nodal_batch_of_trajectories, reconstruction
      )

  @parameterized.parameters(
      dict(integrator=time_integration.crank_nicolson_rk2),
      dict(integrator=time_integration.imex_rk_sil3),
      dict(integrator=time_integration.backward_forward_euler),
  )
  def test_xarray_to_primitive_equations_with_time_data(self, integrator):
    wavenumbers = 21
    layers = 6
    inner_steps = 3
    outer_steps = 12
    physics_specs = primitive_equations.PrimitiveEquationsSpecs.from_si()
    grid = spherical_harmonic.Grid.with_wavenumbers(
        wavenumbers, radius=physics_specs.radius
    )
    vertical_grid = sigma_coordinates.SigmaCoordinates.equidistant(layers)
    coords = coordinate_systems.CoordinateSystem(grid, vertical_grid)
    dt = physics_specs.nondimensionalize(10 * scales.units.minute)

    # creating state, trajectory and batch of trajectories to be converted.
    initial_state_fn, aux_features = primitive_equations_states.steady_state_jw(
        coords, physics_specs
    )
    ref_temperatures = aux_features[xarray_utils.REF_TEMP_KEY]
    nodal_orography = aux_features[xarray_utils.OROGRAPHY]
    orography = primitive_equations.truncated_modal_orography(
        nodal_orography, coords
    )
    state = initial_state_fn()
    state.sim_time = 0.0
    equation = primitive_equations.PrimitiveEquations(
        ref_temperatures, orography, coords, physics_specs
    )
    step_fn = integrator(equation, dt)
    trajectory_fn = jax.jit(
        time_integration.trajectory_from_step(
            step_fn, outer_steps, inner_steps, start_with_input=True
        )
    )
    _, state_trajectory = trajectory_fn(state)
    times = dt * inner_steps * np.arange(outer_steps)
    ds = xarray_utils.data_to_xarray(
        state_trajectory.asdict(), sample_ids=None, times=times, coords=coords
    )
    reconstructed = xarray_utils.xarray_to_primitive_equations_with_time_data(
        ds
    )
    with self.subTest('round_trip'):
      for actual, expected in zip(
          jax.tree_util.tree_leaves(reconstructed),
          jax.tree_util.tree_leaves(state_trajectory.asdict()),
      ):
        np.testing.assert_allclose(actual, expected)
    with self.subTest('simulation_time'):
      np.testing.assert_allclose(ds.time.values, ds.sim_time.values, atol=1e-6)

  def test_xarray_to_data_with_renaming(self):
    wavenumbers = 21
    layers = 6
    physics_specs = primitive_equations.PrimitiveEquationsSpecs.from_si()
    grid = spherical_harmonic.Grid.with_wavenumbers(
        wavenumbers, radius=physics_specs.radius
    )
    vertical_grid = sigma_coordinates.SigmaCoordinates.equidistant(layers)
    coords = coordinate_systems.CoordinateSystem(grid, vertical_grid)
    # creating state to be converted.
    initial_state_fn, _ = primitive_equations_states.steady_state_jw(
        coords, physics_specs
    )
    state = initial_state_fn().asdict()
    state['tracers']['specific_humidity'] = state['divergence']  # mock tracer.
    renaming_dict = {  # dataset names: primitive equation state notation.
        'delta_renamed': 'divergence',
        'nu_renamed': 'vorticity',
        'dt_renamed': 'temperature_variation',
        'lsp_value': 'log_surface_pressure',
        'q_stuff': 'specific_humidity',
    }
    base_to_xarray_fn = xarray_utils.data_to_xarray
    base_from_xarray_fn = functools.partial(
        xarray_utils.xarray_to_primitive_eq_data,
        tracers_to_include=('specific_humidity',),
    )
    ds = xarray_utils.data_to_xarray_with_renaming(
        state,
        to_xarray_fn=base_to_xarray_fn,
        renaming_dict=renaming_dict,
        coords=coords,
        sample_ids=None,
        times=None,
    )
    reconstructed = xarray_utils.xarray_to_data_with_renaming(
        ds, xarray_to_data_fn=base_from_xarray_fn, renaming_dict=renaming_dict
    )
    with self.subTest('dataset_names'):
      self.assertSameElements(ds.keys(), renaming_dict.keys())
    with self.subTest('round_trip'):
      for actual, expected in zip(
          jax.tree_util.tree_leaves(reconstructed),
          jax.tree_util.tree_leaves(state),
      ):
        np.testing.assert_allclose(actual, expected)

  @parameterized.parameters(
      dict(
          coords=coordinate_systems.CoordinateSystem(
              spherical_harmonic.Grid.T21(),
              vertical_interpolation.PressureCoordinates([50, 100, 800]),
          ),
          grid_type='CUBIC',
      ),
      dict(
          coords=coordinate_systems.CoordinateSystem(
              spherical_harmonic.Grid.T42(),
              vertical_interpolation.PressureCoordinates([1, 30, 50, 900]),
          ),
          grid_type='CUBIC',
      ),
      dict(
          coords=coordinate_systems.CoordinateSystem(
              spherical_harmonic.Grid.T21(latitude_spacing='equiangular'),
              vertical_interpolation.PressureCoordinates([10, 1000, 2000]),
          ),
          grid_type='CUBIC',
      ),
      dict(
          coords=coordinate_systems.CoordinateSystem(
              spherical_harmonic.Grid.T85(
                  latitude_spacing='equiangular_with_poles'
              ),
              vertical_interpolation.PressureCoordinates(
                  [10, 50, 500, 600, 900]
              ),
          ),
          grid_type='CUBIC',
      ),
      dict(
          coords=coordinate_systems.CoordinateSystem(
              spherical_harmonic.Grid.TL127(
                  latitude_spacing='equiangular_with_poles'
              ),
              vertical_interpolation.PressureCoordinates(
                  [10, 50, 500, 600, 900]
              ),
          ),
          grid_type='LINEAR',
      ),
      dict(
          coords=coordinate_systems.CoordinateSystem(
              spherical_harmonic.Grid.TL31(),
              vertical_interpolation.PressureCoordinates(
                  [10, 50, 500, 600, 900]
              ),
          ),
          grid_type='LINEAR',
      ),
  )
  def test_coordinate_system_from_dataset(self, coords, grid_type):
    data_array = np.zeros(coords.nodal_shape)
    data_template = {'u': 0, 'v': 0, 'z': 0, 't': 0, 'tracers': {}}
    wb_data = jax.tree_util.tree_map(lambda x: data_array, data_template)
    ds = xarray_utils.data_to_xarray(wb_data, coords=coords, times=None)
    with self.subTest('using_metadata'):
      reconstructed = xarray_utils.coordinate_system_from_dataset(ds)
      self.assertEqual(reconstructed, coords)

    with self.subTest('using_xarray_axes'):
      ds.attrs = {}  # remove metadata.
      reconstructed = xarray_utils.coordinate_system_from_dataset(ds, grid_type)
      self.assertEqual(reconstructed, coords)

  def test_xarray_to_state_and_dynamic_covariate_data(self):
    coords = coordinate_systems.CoordinateSystem(
        spherical_harmonic.Grid.T21(),
        vertical_interpolation.PressureCoordinates([50, 100, 800]),
    )
    times = np.datetime64('1990-01-01') + np.arange(5) * np.timedelta64(1, 'h')

    # Construct xarray dataset from mock state data
    data_array = np.zeros((len(times),) + coords.nodal_shape)
    data_template = {'u': 0, 'v': 0, 'z': 0, 't': 0, 'tracers': {}}
    data = jax.tree_util.tree_map(lambda x: data_array, data_template)
    ds_state = xarray_utils.data_to_xarray(data, coords=coords, times=times)

    # Construct xarray dataset from mock surface covariates
    data_array = np.zeros((len(times),) + coords.surface_nodal_shape)
    data_template = {'sea_ice_cover': 0, 'sea_surface_temperature': 0}
    data = jax.tree_util.tree_map(lambda x: data_array, data_template)
    data['sim_time'] = np.arange(5)  # expect nondim time in prepreocessed data
    ds_surface_covariates = xarray_utils.dynamic_covariate_data_to_xarray(
        data, coords=coords, times=times
    )

    # Construct xarray dataset from mock volume covariates
    data_array = np.zeros((len(times),) + coords.nodal_shape)
    data_template = {'cloud_cover': 0}
    data = jax.tree_util.tree_map(lambda x: data_array, data_template)
    data['sim_time'] = np.arange(5)  # expect nondim time in prepreocessed data
    ds_volume_covariates = xarray_utils.dynamic_covariate_data_to_xarray(
        data, coords=coords, times=times
    )

    ds = xarray.merge([ds_state, ds_surface_covariates, ds_volume_covariates])
    xarray_to_dynamic_covariate_data_fn = functools.partial(
        xarray_utils.xarray_to_dynamic_covariate_data,
        covariates_to_include=['sea_ice_cover', 'cloud_cover'],
    )
    state_data, covariate_data = (
        xarray_utils.xarray_to_state_and_dynamic_covariate_data(
            ds,
            xarray_to_state_data_fn=xarray_utils.xarray_to_weatherbench_data,
            xarray_to_dynamic_covariate_data_fn=(
                xarray_to_dynamic_covariate_data_fn
            ),
        )
    )

    with self.subTest('state data'):
      expected_keys = {'u', 'v', 'z', 't', 'tracers', 'diagnostics', 'sim_time'}
      self.assertSetEqual(set(state_data.keys()), expected_keys)
      sim_time = state_data.pop('sim_time')
      _ = state_data.pop('tracers')
      _ = state_data.pop('diagnostics')
      self.assertEqual(sim_time.shape, (5,))
      for var in state_data.values():
        self.assertEqual(var.shape, (5, 3, 64, 32))

    with self.subTest('covariate data'):
      expected_keys = {'cloud_cover', 'sea_ice_cover', 'sim_time'}
      self.assertSetEqual(set(covariate_data.keys()), expected_keys)
      self.assertEqual(covariate_data['cloud_cover'].shape, (5, 3, 64, 32))
      self.assertEqual(covariate_data['sea_ice_cover'].shape, (5, 1, 64, 32))
      self.assertEqual(covariate_data['sim_time'].shape, (5,))

  @parameterized.parameters(
      dict(time_shift_type='str'),
      dict(time_shift_type='np'),
      dict(time_shift_type='pd'),
  )
  def test_selective_temporal_shift(self, time_shift_type):
    time = np.datetime64('2000-01-01') + [
        np.timedelta64(x, 'h') for x in range(5)
    ]
    input_ds = xarray.Dataset(
        data_vars=dict(
            a=(['time'], [40, 41, 42, 43, 44]),
            b=(['time'], [50, 51, 52, 53, 54]),
            c=(['time'], [60.0, 61.0, 62.0, 63.0, 64.0]),
        ),
        coords=dict(time=time),
    )

    with self.subTest('zero shift'):
      match time_shift_type:
        case 'str':
          time_shift = '0 hours'
        case 'np':
          time_shift = np.timedelta64(0, 'h')
        case 'pd':
          time_shift = pd.Timedelta(0, 'h')
      shifted_ds = xarray_utils.selective_temporal_shift(
          input_ds, variables=['b', 'c'], time_shift=time_shift
      )
      xarray.testing.assert_equal(shifted_ds, input_ds)

    with self.subTest('empty variables list'):
      match time_shift_type:
        case 'str':
          time_shift = '1 hours'
        case 'np':
          time_shift = np.timedelta64(1, 'h')
        case 'pd':
          time_shift = pd.Timedelta(1, 'h')
      shifted_ds = xarray_utils.selective_temporal_shift(
          input_ds, variables=[], time_shift=time_shift
      )
      xarray.testing.assert_equal(shifted_ds, input_ds)

    with self.subTest('positive shift'):
      # Shifting ['b', 'c'] by 1 hour relative to the dataset implies that for
      # a given time they have their value from 1 hour earlier.
      match time_shift_type:
        case 'str':
          time_shift = '1 hour'
        case 'np':
          time_shift = np.timedelta64(1, 'h')
        case 'pd':
          time_shift = pd.Timedelta(1, 'h')
      shifted_ds = xarray_utils.selective_temporal_shift(
          input_ds, variables=['b', 'c'], time_shift=time_shift
      )
      expected_shifted_ds = xarray.Dataset(
          data_vars=dict(
              a=(['time'], [41, 42, 43, 44]),  # unshifted
              b=(['time'], [50, 51, 52, 53]),  # shifted right
              c=(['time'], [60.0, 61.0, 62.0, 63.0]),  # shifted right
          ),
          coords=dict(time=time[1:]),  # remove first entry
      )
      xarray.testing.assert_equal(shifted_ds, expected_shifted_ds)

    with self.subTest('negative shift'):
      # Shifting ['b', 'c'] by -2 hours relative to the dataset implies that for
      # a given time they have their value from 2 hours later.
      match time_shift_type:
        case 'str':
          time_shift = '-2 hours'
        case 'np':
          time_shift = np.timedelta64(-2, 'h')
        case 'pd':
          time_shift = pd.Timedelta(-2, 'h')
      shifted_ds = xarray_utils.selective_temporal_shift(
          input_ds, variables=['b', 'c'], time_shift=time_shift
      )
      expected_shifted_ds = xarray.Dataset(
          data_vars=dict(
              a=(['time'], [40, 41, 42]),  # unshifted
              b=(['time'], [52, 53, 54]),  # shifted left
              c=(['time'], [62.0, 63.0, 64.0]),  # shifted left
          ),
          coords=dict(time=time[:-2]),  # removed last two entries
      )
      xarray.testing.assert_equal(shifted_ds, expected_shifted_ds)

  def test_fill_nan_nearest(self):

    times = np.arange(2)
    rng = np.random.default_rng(seed=0)
    coords = coordinate_systems.CoordinateSystem(
        spherical_harmonic.Grid.T21(),
        vertical_interpolation.PressureCoordinates([50, 100, 800]),
    )

    def to_xarray(data_dict):
      return xarray_utils.data_to_xarray(
          data_dict, coords=coords, times=times
      ).rename({'lon': 'longitude', 'lat': 'latitude'})

    with_valid_nan = rng.random(times.shape + coords.horizontal.nodal_shape)
    with_valid_nan[:, 10:15, 3] = np.nan  # longitude nan slice.
    data_dict = {
        'u': rng.random(times.shape + coords.nodal_shape),
        'v': rng.random(times.shape + coords.nodal_shape),
        'with_nan': with_valid_nan,
    }
    ds_with_nan = to_xarray(data_dict)
    filled_ds = xarray_utils.fill_nan_with_nearest(ds_with_nan)

    with self.subTest('correct_along_longitude'):
      # longitude neighbors are closest on the outside of the nan slice.
      np.testing.assert_allclose(
          filled_ds.with_nan.values[:, 10, 3], with_valid_nan[:, 9, 3]
      )
      np.testing.assert_allclose(
          filled_ds.with_nan.values[:, 14, 3], with_valid_nan[:, 15, 3]
      )

    with self.subTest('correct_along_latitude'):
      # in the center upper latitude neighbor is closest.
      np.testing.assert_allclose(
          filled_ds.with_nan.values[:, 12, 3], with_valid_nan[:, 12, 2]
      )

    with self.subTest('extra_scalar_variable_okay'):
      filled_ds_with_extra = xarray_utils.fill_nan_with_nearest(
          ds_with_nan.assign(sim_time=1.0)
      )
      expected = filled_ds.assign(sim_time=1.0)
      xarray.testing.assert_identical(filled_ds_with_extra, expected)

    with self.subTest('allows_no_time_dimension'):
      filled_ds_no_time = xarray_utils.fill_nan_with_nearest(
          ds_with_nan.isel(time=0)
      )
      xarray.testing.assert_identical(filled_ds_no_time, filled_ds.isel(time=0))

    with self.subTest('raises_on_missing_lat_lon'):
      with self.assertRaisesRegex(
          ValueError, 'did not find latitude and longitude dimensions'
      ):
        xarray_utils.fill_nan_with_nearest(
            ds_with_nan.rename({'latitude': 'lat'})
        )

    with self.subTest('raises_on_non_stationary_mask'):
      with_invalid_nan = rng.random(times.shape + coords.horizontal.nodal_shape)
      with_invalid_nan[:, 10, 3] = np.array([0, np.nan])
      data_dict = {
          'u': rng.random(times.shape + coords.nodal_shape),
          'v': rng.random(times.shape + coords.nodal_shape),
          'with_nan': with_invalid_nan,
      }
      ds_with_invalid_nan = to_xarray(data_dict)
      with self.assertRaisesRegex(ValueError, 'NaN mask is not fixed'):
        xarray_utils.fill_nan_with_nearest(ds_with_invalid_nan)

    with self.subTest('raises_on_all_nan'):
      data_dict = {
          'u': rng.random(times.shape + coords.nodal_shape),
          'v': rng.random(times.shape + coords.nodal_shape),
          'with_nan': with_invalid_nan * np.nan,
      }
      ds_with_invalid_nan = to_xarray(data_dict)
      with self.assertRaisesRegex(ValueError, 'all values are NaN'):
        xarray_utils.fill_nan_with_nearest(ds_with_invalid_nan)

    with self.subTest('works_on_dataarray'):
      filled_u = xarray_utils.fill_nan_with_nearest(ds_with_nan.u)
      xarray.testing.assert_identical(filled_u, filled_ds.u)

  @parameterized.parameters(
      dict(regridder_cls=horizontal_interpolation.BilinearRegridder),
      dict(regridder_cls=horizontal_interpolation.ConservativeRegridder),
      dict(regridder_cls=horizontal_interpolation.NearestRegridder),
  )
  def test_regrid_horizontal(self, regridder_cls):
    old_coords = coordinate_systems.CoordinateSystem(
        spherical_harmonic.Grid.TL31(),
        vertical_interpolation.PressureCoordinates([50, 100, 150]),
    )
    rng = np.random.default_rng(seed=0)
    data_dict = {
        'u': rng.random(old_coords.nodal_shape),
        'v': rng.random(old_coords.surface_nodal_shape),
    }
    ds = (
        xarray_utils.data_to_xarray(data_dict, coords=old_coords, times=None)
        .rename({'lon': 'longitude', 'lat': 'latitude'})
        .squeeze('surface')
    )

    regridder = regridder_cls(
        old_coords.horizontal, spherical_harmonic.Grid.TL63()
    )
    ds_regridded = xarray_utils.regrid_horizontal(ds, regridder)
    expected_sizes = {'latitude': 64, 'longitude': 128, 'level': 3}
    self.assertEqual(ds_regridded.sizes, expected_sizes)

    u_regridded = xarray_utils.regrid_horizontal(ds.u, regridder)
    xarray.testing.assert_identical(u_regridded, ds_regridded.u)

    ds_flipped = ds.isel(latitude=slice(None, None, -1))
    ds_flipped_regridded = xarray_utils.regrid_horizontal(ds_flipped, regridder)
    xarray.testing.assert_identical(ds_flipped_regridded, ds_regridded)

    with self.assertRaisesRegex(ValueError, 'inconsistent latitude'):
      xarray_utils.regrid_horizontal(
          ds.assign(latitude=ds.latitude - 180), regridder
      )

  @parameterized.parameters(
      dict(regridder_cls=vertical_interpolation.BilinearRegridder),
      dict(regridder_cls=vertical_interpolation.ConservativeRegridder),
  )
  def test_regrid_vertical(self, regridder_cls):
    old_coords = vertical_interpolation.HybridCoordinates(
        a_boundaries=np.array([0, 400, 300, 0]),
        b_boundaries=np.array([0, 0, 0.5, 1.0]),
    )
    new_coords = sigma_coordinates.SigmaCoordinates.equidistant(4)

    with self.subTest('single time slice'):
      surface_pressure = xarray.DataArray(
          1000 * np.ones((10, 12)), dims=['longitude', 'latitude']
      )
      ds = xarray.Dataset(
          {'u': (('z', 'longitude', 'latitude'), np.ones((3, 10, 12)))}
      )

      regridder = regridder_cls(old_coords, new_coords)
      ds_regridded = xarray_utils.regrid_vertical(
          ds, surface_pressure, regridder, dim='z'
      )

      expected_sizes = {'sigma': 4, 'longitude': 10, 'latitude': 12}
      self.assertEqual(ds_regridded.sizes, expected_sizes)

      expected_sigma = np.array([0.125, 0.375, 0.625, 0.875])
      np.testing.assert_array_equal(ds_regridded['sigma'].data, expected_sigma)

    with self.subTest('with leading time dimension'):
      surface_pressure = xarray.DataArray(
          1000 * np.ones((2, 10, 12)), dims=['time', 'longitude', 'latitude']
      )
      ds = xarray.Dataset({
          'u': (('time', 'z', 'longitude', 'latitude'), np.ones((2, 3, 10, 12)))
      })
      regridder = regridder_cls(old_coords, new_coords)
      ds_regridded = xarray_utils.regrid_vertical(
          ds, surface_pressure, regridder, dim='z'
      )
      expected_sizes = {'time': 2, 'sigma': 4, 'longitude': 10, 'latitude': 12}
      self.assertEqual(ds_regridded.sizes, expected_sizes)


if __name__ == '__main__':
  jax.config.update('jax_traceback_filtering', 'off')
  absltest.main()



# Copyright 2023 Google LLC

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     https://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for shallow_water_states."""

import operator

from absl.testing import absltest
from absl.testing import parameterized

from dinosaur import scales
from dinosaur import shallow_water_states

import jax
import numpy as np


DEFAULT_SCALE = scales.DEFAULT_SCALE


def assert_array_less_equal(x, y, err_msg='', verbose=True):
  return np.testing.assert_array_compare(
      operator.__le__, x, y, err_msg=err_msg,
      verbose=verbose,
      header='x is not less than or equal to y.',
      equal_inf=False)


class BarotropicInstabilityTest(parameterized.TestCase):

  @parameterized.parameters(
      dict(a=2, b=1, expected=1),
      dict(a=1, b=-1, expected=2),
      dict(a=.1, b=2 * np.pi - .1, expected=.2),
      dict(a=2 * np.pi - .1, b=.1, expected=-.2),
      dict(a=np.pi + 1, b=np.pi - 1, expected=2),
      dict(a=np.pi - 1, b=np.pi + 1, expected=-2),
      dict(a=np.pi, b=np.pi, expected=0)
  )
  def testLongitudeSubtraction(self, a, b, expected):
    np.testing.assert_allclose(
        shallow_water_states.subtract_longitudes(a, b), expected, atol=1e-5)

  @parameterized.parameters(
      dict(seed=0),
      dict(seed=1),
      dict(seed=2)
  )
  def testZonalVelocity(self, seed):
    parameters = shallow_water_states.get_random_parameters(
        jax.random.PRNGKey(seed), shallow_water_states.get_default_parameters())
    parameters = jax.tree.map(DEFAULT_SCALE.nondimensionalize, parameters)

    latitude = np.linspace(-np.pi / 2, np.pi / 2, 101)
    zonal_velocity = shallow_water_states.get_zonal_velocity(
        latitude, parameters)
    atol = np.finfo(np.float32).eps

    with self.subTest('IsZeroOutsideJet'):
      outside_jet = ((latitude > parameters.jet_northern_lat)
                     + (latitude < parameters.jet_southern_lat))
      np.testing.assert_allclose(outside_jet * zonal_velocity, 0)

    with self.subTest('IsBoundedByMaxVelocity'):
      assert_array_less_equal(
          np.abs(zonal_velocity), parameters.jet_max_velocity + atol
      )

    with self.subTest('AttainsMaxVelocityAtJetCenter'):
      center = (parameters.jet_northern_lat + parameters.jet_southern_lat) / 2
      center_velocity = shallow_water_states.get_zonal_velocity(
          center, parameters)
      np.testing.assert_allclose(
          center_velocity, parameters.jet_max_velocity, atol=atol
      )

  @parameterized.parameters(
      dict(seed=0),
      dict(seed=1),
      dict(seed=2)
  )
  def testHeight(self, seed):
    parameters = shallow_water_states.get_random_parameters(
        jax.random.PRNGKey(seed), shallow_water_states.get_default_parameters())
    parameters = jax.tree.map(DEFAULT_SCALE.nondimensionalize, parameters)
    longitude = np.linspace(0, 2 * np.pi, 101)
    latitude = np.linspace(-np.pi / 2, np.pi / 2, 101)
    longitude, latitude = np.meshgrid(longitude, latitude, indexing='ij')
    height = shallow_water_states.get_height(longitude, latitude, parameters)

    with self.subTest('BumpInsideJet'):
      self.assertGreaterEqual(parameters.bump_lat_location,
                              parameters.jet_southern_lat)
      self.assertLessEqual(parameters.bump_lat_location,
                           parameters.jet_northern_lat)

    with self.subTest('IsNonNegative'):
      assert_array_less_equal(0, height)

    with self.subTest('IsBoundedByScale'):
      assert_array_less_equal(height, parameters.bump_height_scale)

    with self.subTest('AttainsMaxHeightAtCenter'):
      max_height = (
          np.cos(parameters.bump_lat_location) * parameters.bump_height_scale)
      center_height = shallow_water_states.get_height(
          parameters.bump_lon_location,
          parameters.bump_lat_location,
          parameters)
      np.testing.assert_allclose(center_height, max_height)

if __name__ == '__main__':
  absltest.main()



# Copyright 2023 Google LLC

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     https://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Real Fourier basis evaluation, and derivative coefficients."""
from dinosaur import jax_numpy_utils

import jax
import jax.numpy as jnp
import numpy as np
import scipy.linalg


def real_basis(wavenumbers: int, nodes: int) -> np.ndarray:
  """Returns the real-valued Fourier basis.

  Args:
    wavenumbers: number of wavenumbers.
    nodes: number of equally spaced nodes in the range [0, 2π). Must satisfy
      wavenumbers >= nodes.

  Returns:
    The nodes ✕ (2 * wavenumbers - 1) matrix F, such that

      F[i, 0] = 1 / √2π
      F[i, 2j - 1] = cos(j xᵢ) / √π    1 ≤ j < wavenumbers
      F[i, 2j]     = sin(j xᵢ) / √π    1 ≤ j < wavenumbers

    where x is a vector of m equally-spaced nodes so xᵢ = i * 2π / nodes.
    i.e., the columns of F are the real Fourier basis functions sampled at x.

    The normalization of the basis functions is chosen such that they have unit
    L²([0, 2π]) norm.
  """
  if nodes < wavenumbers:
    raise ValueError(
        '`real_basis` requires nodes >= wavenumbers; '
        f'got m = {nodes} and n = {wavenumbers}.'
    )

  dft = scipy.linalg.dft(nodes)[:, :wavenumbers] / np.sqrt(np.pi)

  cos = np.real(dft[:, 1:])
  sin = -np.imag(dft[:, 1:])

  f = np.empty(shape=[nodes, 2 * wavenumbers - 1], dtype=np.float64)
  f[:, 0] = 1 / np.sqrt(2 * np.pi)
  f[:, 1::2] = cos
  f[:, 2::2] = sin
  return f


def real_basis_derivative(
    u: jnp.ndarray | jax.Array, /, axis: int = -1
) -> jax.Array:
  """Calculate the derivative of a signal using a real basis.

  Args:
    u: signal to differentiate, in the real Fourier basis.
    axis: the axis along which the transform will be applied.

  Returns:
    The derivative of `u` along `axis`. In particular, if
    `u_x = real_basis_derivative(u)`:

      u_x[..., 0]      = 0
      u_x[..., 2j - 1] =  j * u[..., 2j]      1 ≤ j < wavenumbers
      u_x[..., 2j]     = -j * u[..., 2j - 1]  1 ≤ j < wavenumbers
  """
  if u.shape[axis] % 2 != 1:
    raise ValueError(f'{u.shape=} along {axis=} is not odd')
  if axis >= 0:
    raise ValueError('axis must be negative')
  i = jnp.arange(u.shape[axis]).reshape((-1,) + (1,) * (-1 - axis))
  j = (i + 1) // 2
  u_down = jax_numpy_utils.shift(u, -1, axis)
  u_up = jax_numpy_utils.shift(u, +1, axis)
  return j * jnp.where(i % 2, u_down, -u_up)


def real_basis_with_zero_imag(wavenumbers: int, nodes: int) -> np.ndarray:
  """Real basis with a zero imaginary part."""
  if nodes < wavenumbers:
    raise ValueError(
        '`real_basis` requires nodes >= wavenumbers; '
        f'got m = {nodes} and n = {wavenumbers}.'
    )

  dft = scipy.linalg.dft(nodes)[:, :wavenumbers] / np.sqrt(np.pi)

  cos = np.real(dft[:, 1:])
  sin = -np.imag(dft[:, 1:])

  f = np.empty(shape=[nodes, 2 * wavenumbers], dtype=np.float64)
  f[:, 0] = 1 / np.sqrt(2 * np.pi)
  f[:, 1] = 0
  f[:, 2::2] = cos
  f[:, 3::2] = sin
  return f


def real_basis_derivative_with_zero_imag(
    u: jnp.ndarray | jax.Array, axis: int = -1, frequency_offset: int = 0
) -> jax.Array:
  """Calculate the derivative along a real basis with zero imaginary part."""
  if u.shape[axis] % 2:
    raise ValueError(f'{u.shape=} along {axis=} is not even')
  if axis >= 0:
    raise ValueError('axis must be negative')
  i = jnp.arange(u.shape[axis]).reshape((-1,) + (1,) * (-1 - axis))
  j = frequency_offset + i // 2
  u_down = jax_numpy_utils.shift(u, -1, axis)
  u_up = jax_numpy_utils.shift(u, +1, axis)
  return j * jnp.where((i + 1) % 2, u_down, -u_up)


def quadrature_nodes(nodes: int) -> tuple[np.ndarray, np.ndarray]:
  """Returns nodes and weights for the trapezoidal rule.

  Args:
    nodes: the number of nodes and weights to return.

  Returns:
    A pair (nodes, weights); values and weights to use for quadrature, with
    dtype `np.float64`.
  """
  xs = np.linspace(0, 2 * np.pi, nodes, endpoint=False)
  weights = 2 * np.pi / nodes
  return xs, weights



# Copyright 2023 Google LLC

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     https://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Routines for regridding between sigma and pressure levels."""
from __future__ import annotations

import dataclasses
import functools
import importlib
from typing import Any, Callable, Dict, Sequence, Union

import dinosaur
from dinosaur import pytree_utils
from dinosaur import sigma_coordinates
from dinosaur import typing
import jax
import jax.numpy as jnp
import numpy as np


Array = typing.Array
InterpolateFn = Callable[[Array, Array, Array], Array]


def vectorize_vertical_interpolation(
    interpolate_fn: InterpolateFn,
) -> InterpolateFn:
  """Vectorizes vertical `interpolate_fn` function to work on 3+d fields."""
  # inerpolate_fn operates on `x, xp, fp` (target, source loc, source vals).
  # vmap over spatial axes for targets and source values.
  interpolate_fn = jax.vmap(interpolate_fn, (-1, None, -1), out_axes=-1)
  interpolate_fn = jax.vmap(interpolate_fn, (-1, None, -1), out_axes=-1)
  # vmap over multiple vertical targets.
  interpolate_fn = jax.vmap(interpolate_fn, (0, None, None), out_axes=0)
  # vectorize an leading dimensions.
  return jnp.vectorize(interpolate_fn, signature='(a,x,y),(b),(b,x,y)->(a,x,y)')


def _dot_interp(x, xp, fp):
  # pylint: disable=g-doc-return-or-yield,g-doc-args
  """Interpolate with a dot product instead of indexing.

  This is often much faster than the default jnp.interp on TPUs.
  """
  # TODO(shoyer): upstream this into jnp.interp
  n = len(xp)
  i = jnp.arange(n)
  dx = xp[1:] - xp[:-1]
  delta = x - xp[:-1]
  w = delta / dx
  w_left = jnp.pad(1 - w, [(0, 1)])
  w_right = jnp.pad(w, [(1, 0)])
  u = jnp.searchsorted(xp, x, side='right', method='compare_all')
  u = jnp.clip(u, 1, n - 1)
  weights = w_left * (i == (u - 1)) + w_right * (i == u)
  weights = jnp.where(x < xp[0], i == 0, weights)
  weights = jnp.where(x > xp[-1], i == (n - 1), weights)
  return jnp.dot(weights, fp, precision='highest')


@jax.jit
def interp(
    x: typing.Numeric, xp: typing.Array, fp: typing.Array
) -> jnp.ndarray:
  """Optimized version of jnp.interp."""
  return jax.lax.platform_dependent(
      x, xp, fp, tpu=_dot_interp, default=jnp.interp
  )


def _extrapolate_left(y):
  delta = y[1] - y[0]
  return jnp.concatenate([jnp.array([y[0] - delta]), y])


def _extrapolate_right(y):
  delta = y[-1] - y[-2]
  return jnp.concatenate([y, jnp.array([y[-1] + delta])])


def _extrapolate_both(y):
  return _extrapolate_left(_extrapolate_right(y))


def _linear_interp_with_safe_extrap(x, xp, fp, n=1):
  """Linear interpolation with extrapolation for n grid cells at each end."""
  for _ in range(n):
    xp = _extrapolate_both(xp)
    fp = _extrapolate_both(fp)
  return jnp.interp(x, xp, fp, left=np.nan, right=np.nan)


@jax.jit
def linear_interp_with_linear_extrap(
    x: typing.Numeric, xp: typing.Array, fp: typing.Array
) -> jnp.ndarray:
  """Linear interpolation with unlimited linear extrapolation at each end."""
  n = len(xp)
  i = jnp.arange(n)
  dx = xp[1:] - xp[:-1]
  delta = x - xp[:-1]
  w = delta / dx
  w_left = jnp.pad(1 - w, [(0, 1)])
  w_right = jnp.pad(w, [(1, 0)])
  u = jnp.searchsorted(xp, x, side='right', method='compare_all')
  u = jnp.clip(u, 1, n - 1)
  weights = w_left * (i == (u - 1)) + w_right * (i == u)
  return jnp.dot(weights, fp, precision='highest')


# TODO(shoyer): add higher order interpolation schemes, e.g., with splines.
# These aren't supported in JAX yet, but can be found in JAX-Cosmo:
# https://github.com/DifferentiableUniverseInitiative/jax_cosmo/issues/29
# https://github.com/DifferentiableUniverseInitiative/jax_cosmo/blob/master/jax_cosmo/scipy/interpolate.py


@dataclasses.dataclass(frozen=True)
class PressureCoordinates:
  """Specifies the vertical coordinate with pressure levels.

  Attributes:
    centers: center of each pressure level, starting at the level closest to the
      top of the atmosphere. Must be monotonic increasing.
    layers: number of vertical layers.
  """

  centers: np.ndarray

  def __init__(self, centers: Union[Sequence[float], np.ndarray]):
    object.__setattr__(self, 'centers', np.asarray(centers))
    if not all(np.diff(self.centers) > 0):
      raise ValueError(
          'Expected `centers` to be monotonically increasing, '
          f'got centers = {self.centers}'
      )

  @property
  def layers(self) -> int:
    return len(self.centers)

  def asdict(self) -> Dict[str, Any]:
    return {k: v.tolist() for k, v in dataclasses.asdict(self).items()}

  def __hash__(self):
    return hash(tuple(self.centers.tolist()))

  def __eq__(self, other):
    return isinstance(other, PressureCoordinates) and np.array_equal(
        self.centers, other.centers
    )


@dataclasses.dataclass(frozen=True)
class HybridCoordinates:
  """Specifies the vertical coordinate with hybrid sigma-pressure levels.

  This allows for matching the vertical coordinate system used by ECMWF and most
  other operational forecasting systems.

  The pressure corresponding to a level is given by the formula `a + b * sp`
  where `sp` is surface pressure.

  Attributes:
    a_boundaries: offset coefficient for the boundaries of each level, starting
      at the level closest to the top of the atmosphere.
    b_boundaries: slope coefficient for the boundaries of each level, starting
      at the level closest to the top of the atmosphere.
    layers: number of vertical layers.
  """

  a_boundaries: np.ndarray
  b_boundaries: np.ndarray

  def __post_init__(self):
    if len(self.a_boundaries) != len(self.b_boundaries):
      raise ValueError(
          'Expected `a_boundaries` and `b_boundaries` to have the same length, '
          f'got {len(self.a_boundaries)} and {len(self.b_boundaries)}.'
      )

  @classmethod
  def _from_resource_csv(cls, path: str) -> HybridCoordinates:
    levels_csv = importlib.resources.files(dinosaur).joinpath(path)
    with levels_csv.open() as f:
      a_in_pa, b = np.loadtxt(f, skiprows=1, usecols=(1, 2), delimiter='\t').T
    a = a_in_pa / 100  # convert from Pa to hPa
    # any reasonable hybrid coordinate system falls in this range (certainly
    # including UFS and ECMWF)
    assert 100 < a.max() < 1000
    return cls(a_boundaries=a, b_boundaries=b)

  @classmethod
  def ECMWF137(cls) -> HybridCoordinates:  # pylint: disable=invalid-name
    """Returns the 137 model levels used by ECMWF's IFS (e.g., in ERA5).

    Pressure is returned in units of hPa.

    For details, see the ECMWF wiki:
    https://confluence.ecmwf.int/display/UDOC/L137+model+level+definitions
    """
    return cls._from_resource_csv('data/ecmwf137_hybrid_levels.csv')

  @classmethod
  def UFS127(cls) -> HybridCoordinates:  # pylint: disable=invalid-name
    """Returns the 127 model levels used by NOAA's UFS (GFS v16).

    Pressure is returned in units of hPa.

    For details, see the documentation for UFS Replay:
    https://ufs2arco.readthedocs.io/en/latest/example_pressure_interpolation.html

    Source data:
    https://github.com/NOAA-PSL/ufs2arco/blob/v0.1.2/ufs2arco/replay_vertical_levels.yaml
    """
    return cls._from_resource_csv('data/ufs127_hybrid_levels.csv')

  @property
  def layers(self) -> int:
    return len(self.a_boundaries) - 1

  def __hash__(self):
    return hash(
        (tuple(self.a_boundaries.tolist()), tuple(self.b_boundaries.tolist()))
    )

  def __eq__(self, other):
    return (
        isinstance(other, HybridCoordinates)
        and np.array_equal(self.a_boundaries, other.a_boundaries)
        and np.array_equal(self.b_boundaries, other.b_boundaries)
    )

  def get_sigma_boundaries(
      self, surface_pressure: typing.Numeric
  ) -> typing.Array:
    """Returns centers of sigma levels for a given surface pressure.

    Args:
      surface_pressure: scalar surface pressure, in the same units as
        `a_boundaries`.

    Returns:
      Array with shape `(layers + 1,)`.
    """
    return self.a_boundaries / surface_pressure + self.b_boundaries

  def get_sigma_centers(self, surface_pressure: typing.Numeric) -> typing.Array:
    """Returns centers of sigma levels for a given surface pressure.

    Args:
      surface_pressure: scalar surface pressure, in the same units as
        `a_boundaries`.

    Returns:
      Array with shape `(layers,)`.
    """
    boundaries = self.get_sigma_boundaries(surface_pressure)
    return (boundaries[1:] + boundaries[:-1]) / 2

  def to_approx_sigma_coords(
      self, layers: int, surface_pressure: float = 1013.25
  ) -> sigma_coordinates.SigmaCoordinates:
    """Interpolate these hybrid coordinates to approximate sigma levels.

    The resulting coordinates will typically not be equidistant.

    Args:
      layers: number of sigma layers to return.
      surface_pressure: reference surface pressure to use for interpolation. The
        default value is 1013.25, which is one standard atmosphere in hPa.

    Returns:
      New SigmaCoordinates object wih the requested number of layers.
    """
    original_bounds = self.get_sigma_boundaries(surface_pressure)
    interpolated_bounds = jax.vmap(jnp.interp, (0, None, None))(
        jnp.linspace(0, 1, num=layers + 1),
        jnp.linspace(0, 1, num=self.layers + 1),
        original_bounds,
    )
    interpolated_bounds = np.array(interpolated_bounds)
    # Some hybrid coordinates (e.g., from UFS) start at a non-zero pressure
    # level. It is not clear that this makes sense for Dinosaur, so to be safe,
    # we set the first level to 0 (zero pressure) and the last level to 1
    # (surface pressure).
    interpolated_bounds[0] = 0.0
    interpolated_bounds[-1] = 1.0
    return sigma_coordinates.SigmaCoordinates(interpolated_bounds)


@functools.partial(jax.jit, static_argnums=0)
def get_surface_pressure(
    pressure_levels: PressureCoordinates,
    geopotential: typing.Array,
    orography: typing.Array,
    gravity_acceleration: float,
) -> typing.Array:
  """Calculate surface pressure from geopotential on pressure levels.

  Args:
    pressure_levels: 1D array with pressure levels.
    geopotential: array with dimensions [..., level, x, y]
    orography: array with dimensions [1, x, y]
    gravity_acceleration: acceleration due to gravity.

  Returns:
    Array with dimensions [..., x, y].
  """
  # note: relative height must be an increasing function along the level axis,
  # which is why we subtract geopotential (which decreases as you get closer to
  # the surface of the Earth)
  relative_height = orography * gravity_acceleration - geopotential

  @functools.partial(jnp.vectorize, signature='(z,x,y),(z)->(1,x,y)')
  @functools.partial(jax.vmap, in_axes=(-1, None), out_axes=-1)
  @functools.partial(jax.vmap, in_axes=(-1, None), out_axes=-1)
  def find_intercept(rh, levels):
    return linear_interp_with_linear_extrap(0.0, rh, levels)[np.newaxis]

  return find_intercept(relative_height, pressure_levels.centers)


def vertical_interpolation(
    x: typing.Array,
    xp: typing.Array,
    fp: typing.Array,
) -> typing.Array:
  """Computes linear interpolation `f(x)` with constant extrapolation.

  Args:
    x: array of vertical coordinates for which interpolation is computed.
    xp: array of vertical coordinates on which values are known.
    fp: array of values of on the grid `xp`.

  Returns:
    Interpolated values `f(x)`.
  """
  return interp(x, jnp.asarray(xp), fp)


@functools.partial(jax.jit, static_argnums=(1, 2, 4))
def interp_pressure_to_sigma(
    fields: typing.Pytree,
    pressure_coords: PressureCoordinates,
    sigma_coords: sigma_coordinates.SigmaCoordinates,
    surface_pressure: typing.Array,
    interpolate_fn: InterpolateFn = (
        vectorize_vertical_interpolation(_linear_interp_with_safe_extrap)
    ),
) -> typing.Pytree:
  """Interpolate 3D fields from pressure to sigma levels."""
  desired = sigma_coords.centers[:, np.newaxis, np.newaxis] * surface_pressure
  regrid = lambda x: interpolate_fn(desired, pressure_coords.centers, x)

  def cond_fn(x) -> bool:
    shape = jnp.shape(x)
    return len(shape) >= 3 and shape[-3] == pressure_coords.centers.shape[0]

  return pytree_utils.tree_map_where(
      condition_fn=cond_fn,
      f=regrid,
      g=lambda x: x,
      x=fields,
  )


@functools.partial(jax.jit, static_argnums=(1, 2, 4))
def interp_sigma_to_pressure(
    fields: typing.Pytree,
    pressure_coords: PressureCoordinates,
    sigma_coords: sigma_coordinates.SigmaCoordinates,
    surface_pressure: typing.Array,
    interpolate_fn: InterpolateFn = (
        vectorize_vertical_interpolation(_linear_interp_with_safe_extrap)
    ),
) -> typing.Pytree:
  """Interpolate 3D fields from sigma to pressure levels."""
  desired = (
      pressure_coords.centers[:, np.newaxis, np.newaxis] / surface_pressure
  )
  regrid = lambda x: interpolate_fn(desired, sigma_coords.centers, x)
  return pytree_utils.tree_map_over_nonscalars(regrid, fields)


@functools.partial(jax.jit, static_argnums=(1, 2))
def interp_hybrid_to_sigma(
    fields: typing.Pytree,
    hybrid_coords: HybridCoordinates,
    sigma_coords: sigma_coordinates.SigmaCoordinates,
    surface_pressure: typing.Array,
) -> typing.Pytree:
  """Interpolate 3D fields from hybrid to sigma levels."""
  # Linear interpolation from values at cell centers on hybrid levels to cell
  # centers on sigma levels. Here we interpolate on sigma coordinates.

  @jax.jit
  @functools.partial(jnp.vectorize, signature='(x,y),(a),(b,x,y)->(a,x,y)')
  @functools.partial(jax.vmap, in_axes=(-1, None, -1), out_axes=-1)
  @functools.partial(jax.vmap, in_axes=(-1, None, -1), out_axes=-1)
  def regrid(surface_pressure, target_sigmas, field):
    source_sigmas = hybrid_coords.get_sigma_centers(surface_pressure)
    return jax.vmap(_linear_interp_with_safe_extrap, in_axes=(0, None, None))(
        target_sigmas, source_sigmas, field
    )

  return pytree_utils.tree_map_over_nonscalars(
      lambda x: regrid(surface_pressure, sigma_coords.centers, x), fields
  )


def _interval_overlap(
    source_bounds: typing.Array, target_bounds: typing.Array
) -> jnp.ndarray:
  """Calculate the interval overlap between grid cells."""
  # based on https://gist.github.com/shoyer/c0f1ddf409667650a076c058f9a17276
  # (see also horizontal_interpolation.py)
  upper = jnp.minimum(
      target_bounds[1:, jnp.newaxis], source_bounds[jnp.newaxis, 1:]
  )
  lower = jnp.maximum(
      target_bounds[:-1, jnp.newaxis], source_bounds[jnp.newaxis, :-1]
  )
  return jnp.maximum(upper - lower, 0)


def conservative_regrid_weights(
    source_bounds: typing.Array, target_bounds: typing.Array
) -> jnp.ndarray:
  """Create a weight matrix for conservative regridding on pressure levels.

  Args:
    source_bounds: boundaries between increasing pressure levels for the source
      grid. Values must be strictly increasing.
    target_bounds: 1D strictly increasing pressure levels for the target grid.
      Values must be strictly increasing.

  Returns:
    NumPy array with shape (target, source). Rows sum to 1.
  """
  weights = _interval_overlap(source_bounds, target_bounds)
  weights /= jnp.sum(weights, axis=1, keepdims=True)
  assert weights.shape == (target_bounds.size - 1, source_bounds.size - 1)
  return weights


@functools.partial(jax.jit, static_argnums=(1, 2))
def regrid_hybrid_to_sigma(
    fields: typing.Pytree,
    hybrid_coords: HybridCoordinates,
    sigma_coords: sigma_coordinates.SigmaCoordinates,
    surface_pressure: typing.Array,
) -> typing.Pytree:
  """Conservatively regrid 3D fields from hybrid to sigma levels."""
  # Conservative regridding is a simple area-weighted average of source cells.
  # Here we are regridding in one dimension, so it only depends on the overlap
  # between cell boundaries. Here we calculate both bounds in terms of sigma
  # coordinates (from 0 to 1).

  # Note: c=a-1, but the signature argument does not support arithemtic in
  # dimension specifications. This is verified via assert statements below.
  @jax.jit
  @functools.partial(jnp.vectorize, signature='(x,y),(a),(b,x,y)->(c,x,y)')
  @functools.partial(jax.vmap, in_axes=(-1, None, -1), out_axes=-1)
  @functools.partial(jax.vmap, in_axes=(-1, None, -1), out_axes=-1)
  def regrid(surface_pressure, sigma_bounds, field):
    assert sigma_bounds.shape == (sigma_coords.layers + 1,)
    if field.shape[0] != hybrid_coords.layers:
      raise ValueError(
          f'Source has {hybrid_coords.layers} layers, but field has'
          f' {fields.shape[0]}'
      )
    hybrid_bounds = hybrid_coords.get_sigma_boundaries(surface_pressure)
    weights = conservative_regrid_weights(hybrid_bounds, sigma_bounds)
    result = jnp.einsum('ab,b->a', weights, field, precision='float32')
    assert result.shape[0] == sigma_coords.layers
    return result

  return pytree_utils.tree_map_over_nonscalars(
      lambda x: regrid(surface_pressure, sigma_coords.boundaries, x), fields
  )


@dataclasses.dataclass(frozen=True)
class Regridder:
  """Regrid vertically, from hybrid to sigma coordinates."""

  # TODO(shoyer): support more generic vertical coordinate systems.
  source_grid: HybridCoordinates
  target_grid: sigma_coordinates.SigmaCoordinates

  def __call__(
      self, field: typing.Array, surface_pressure: typing.Array
  ) -> jnp.ndarray:
    raise NotImplementedError


@dataclasses.dataclass(frozen=True)
class ConservativeRegridder(Regridder):
  """Regrid with conservative interpolation."""

  def __call__(
      self, field: typing.Array, surface_pressure: typing.Array
  ) -> jnp.ndarray:
    return regrid_hybrid_to_sigma(
        field, self.source_grid, self.target_grid, surface_pressure
    )


@dataclasses.dataclass(frozen=True)
class BilinearRegridder(Regridder):
  """Regrid with bilinear interpolation."""

  def __call__(
      self, field: typing.Array, surface_pressure: typing.Array
  ) -> jnp.ndarray:
    return interp_hybrid_to_sigma(
        field, self.source_grid, self.target_grid, surface_pressure
    )



# Copyright 2023 Google LLC

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     https://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for held_suarez."""

from absl.testing import absltest
from absl.testing import parameterized

from dinosaur import coordinate_systems
from dinosaur import held_suarez
from dinosaur import primitive_equations
from dinosaur import primitive_equations_states
from dinosaur import scales
from dinosaur import sigma_coordinates
from dinosaur import spherical_harmonic
from dinosaur import xarray_utils

import jax
import numpy as np

jax.config.parse_flags_with_absl()


def _subtract_longitudes(a, b):
  """Computes the difference `a - b`, accounting for periodicity."""
  return np.mod(a - b + np.pi, 2 * np.pi) - np.pi


def _gaussian_blob(
    longitudes: np.ndarray,
    latitudes: np.ndarray,
    lon_loc: float = np.pi * 120 / 180,
    lat_loc: float = np.pi * 30 / 180,
    lon_scale: float = np.pi * 10 / 180,
    lat_scale: float = np.pi * 10 / 180,
):
  return np.cos(latitudes) * np.exp(
      -(_subtract_longitudes(lon_loc, longitudes) / lon_scale) ** 2
      -((lat_loc - latitudes) / lat_scale) ** 2
  )


def get_test_orography(
    grid: spherical_harmonic.Grid,
    min_height_in_meters: float = -100.0,
    max_height_in_meters: float = 5000.0,
) -> np.ndarray:
  """Returns random orography with values in the same range as data."""
  lons, sin_lats = grid.nodal_mesh
  lats = np.arcsin(sin_lats)
  return (
      min_height_in_meters * _gaussian_blob(lons, lats, lon_loc=0.0) +
      max_height_in_meters * _gaussian_blob(lons, lats, lat_loc=-np.pi / 6))


class HeldSuarezTest(parameterized.TestCase):

  def test_isothermal_rest_atmosphere_no_orography(self):
    units = scales.units
    layers = 26
    coords = coordinate_systems.CoordinateSystem(
        horizontal=spherical_harmonic.Grid.T42(),
        vertical=sigma_coordinates.SigmaCoordinates.equidistant(layers))
    physics_specs = primitive_equations.PrimitiveEquationsSpecs.from_si()

    initial_state_fn, aux_features = (
        primitive_equations_states.isothermal_rest_atmosphere(
            coords,
            physics_specs,
            p0=1e5 * units.pascal,
            p1=5e3 * units.pascal,
        ))

    ref_temps = aux_features[xarray_utils.REF_TEMP_KEY]
    orography = aux_features[xarray_utils.OROGRAPHY]
    state = initial_state_fn(rng_key=jax.random.PRNGKey(0))

    with self.subTest('state'):
      np.testing.assert_allclose(state.vorticity, 0)
      np.testing.assert_allclose(state.divergence, 0)
      np.testing.assert_allclose(state.temperature_variation, 0)
      self.assertEqual(state.vorticity.shape, coords.modal_shape)
      self.assertEqual(state.divergence.shape, coords.modal_shape)
      self.assertEqual(state.temperature_variation.shape, coords.modal_shape)
      self.assertEqual(state.log_surface_pressure.shape,
                       coords.surface_modal_shape)

    with self.subTest('surface pressure'):
      surface_pressure = physics_specs.dimensionalize(np.exp(
          coords.horizontal.to_nodal(state.log_surface_pressure)), units.pascal)
      np.testing.assert_array_less(surface_pressure, 102000)
      np.testing.assert_array_less(98000, surface_pressure)

    with self.subTest('orography'):
      np.testing.assert_allclose(orography, 0)
      self.assertEqual(orography.shape, coords.horizontal.nodal_shape)

    with self.subTest('reference temperature'):
      np.testing.assert_allclose(ref_temps, 288)
      self.assertEqual(ref_temps.shape, (coords.vertical.layers,))

  def test_isothermal_rest_atmosphere_with_orography(self):
    units = scales.units
    layers = 26
    coords = coordinate_systems.CoordinateSystem(
        horizontal=spherical_harmonic.Grid.T85(),
        vertical=sigma_coordinates.SigmaCoordinates.equidistant(layers))
    physics_specs = primitive_equations.PrimitiveEquationsSpecs.from_si()

    # WeatherBench data is on an equiangular grid.
    orography = get_test_orography(coords.horizontal)

    initial_state_fn, aux_features = (
        primitive_equations_states.isothermal_rest_atmosphere(
            coords,
            physics_specs,
            p0=1e5 * units.pascal,
            p1=5e3 * units.pascal,
            surface_height=orography * units.meter,
        ))

    ref_temps = aux_features[xarray_utils.REF_TEMP_KEY]
    orography = aux_features[xarray_utils.OROGRAPHY]
    state = initial_state_fn(rng_key=jax.random.PRNGKey(0))

    with self.subTest('state'):
      np.testing.assert_allclose(state.vorticity, 0)
      np.testing.assert_allclose(state.divergence, 0)
      np.testing.assert_allclose(state.temperature_variation, 0)
      self.assertEqual(state.vorticity.shape, coords.modal_shape)
      self.assertEqual(state.divergence.shape, coords.modal_shape)
      self.assertEqual(state.temperature_variation.shape, coords.modal_shape)
      self.assertEqual(state.log_surface_pressure.shape,
                       coords.surface_modal_shape)

    with self.subTest('aux_features'):
      for v in aux_features.values():
        self.assertIsInstance(v, np.ndarray)

    with self.subTest('surface pressure'):
      surface_pressure_pa = physics_specs.dimensionalize(np.exp(
          coords.horizontal.to_nodal(state.log_surface_pressure)), units.pascal)
      np.testing.assert_array_less(surface_pressure_pa, 108_000)
      np.testing.assert_array_less(45_000, surface_pressure_pa)

    with self.subTest('orography'):
      self.assertEqual(orography.shape, coords.horizontal.nodal_shape)
      nodal_orography_m = physics_specs.dimensionalize(
          orography, units.meter).magnitude
      np.testing.assert_array_less(nodal_orography_m, 6_000)
      np.testing.assert_array_less(-500, nodal_orography_m)

    with self.subTest('reference temperature'):
      np.testing.assert_allclose(ref_temps, 288)
      self.assertEqual(ref_temps.shape, (coords.vertical.layers,))

  def test_held_suarez_forcing(self):
    units = scales.units
    layers = 26
    coords = coordinate_systems.CoordinateSystem(
        horizontal=spherical_harmonic.Grid.T42(),
        vertical=sigma_coordinates.SigmaCoordinates.equidistant(layers))
    physics_specs = primitive_equations.PrimitiveEquationsSpecs.from_si()

    initial_state_fn, aux_features = (
        primitive_equations_states.isothermal_rest_atmosphere(
            coords, physics_specs, p0=1e5 * units.pascal,
            p1=5e3 * units.pascal,))
    ref_temps = aux_features[xarray_utils.REF_TEMP_KEY]
    state = initial_state_fn(rng_key=jax.random.PRNGKey(0))

    hs = held_suarez.HeldSuarezForcing(
        coords=coords,
        physics_specs=physics_specs,
        reference_temperature=ref_temps)

    self.assertEqual(hs.kv().shape, (coords.vertical.layers, 1, 1))
    self.assertEqual(hs.kt().shape, coords.nodal_shape)

    surface_pressure = np.ones(coords.nodal_shape)
    self.assertEqual(hs.equilibrium_temperature(surface_pressure).shape,
                     coords.nodal_shape)

    explicit_terms = hs.explicit_terms(state)
    np.testing.assert_allclose(explicit_terms.log_surface_pressure, 0)


if __name__ == '__main__':
  absltest.main()
